"""
Using Gammapy IRFs
==================

`gammapy.irf` contains classes for handling Instrument Response
Functions typically stored as multi-dimensional tables. Gammapy is currently supporting
the functions defined in the GADF format (see
https://gamma-astro-data-formats.readthedocs.io/en/v0.3/irfs/full_enclosure/index.html). The
detailed list can be found in the :doc:`IRF user guide </user-guide/irf/index>`.

This tutorial is intended for advanced users typically creating IRFs.

"""

import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.visualization import quantity_support
import matplotlib.pyplot as plt
from gammapy.irf import (
    IRF,
    Background3D,
    EffectiveAreaTable2D,
    EnergyDependentMultiGaussPSF,
    EnergyDispersion2D,
)
from gammapy.irf.io import COMMON_IRF_HEADERS, IRF_DL3_HDU_SPECIFICATION
from gammapy.makers.utils import (
    make_edisp_kernel_map,
    make_map_exposure_true_energy,
    make_psf_map,
)
from gammapy.maps import MapAxis, WcsGeom

######################################################################
# Inbuilt Gammapy IRFs
# --------------------
#

irf_filename = (
    "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
)

aeff = EffectiveAreaTable2D.read(irf_filename, hdu="EFFECTIVE AREA")
print(aeff)


######################################################################
# We can see that the Effective Area Table is defined in terms of
# ``energy_true`` and ``offset`` from the camera center
#

# To see the IRF axes binning, eg, offset
print(aeff.axes["offset"])

# To get the IRF data
print(aeff.data)

# the aeff is evaluated at a given energy and offset
print(aeff.evaluate(energy_true=[1, 10] * u.TeV, offset=[0.2, 2.5] * u.deg))


# The peek method gives a quick look into the IRF
aeff.peek()


######################################################################
# Similarly, we can access other IRFs as well
#

psf = EnergyDependentMultiGaussPSF.read(irf_filename, hdu="Point Spread Function")
bkg = Background3D.read(irf_filename, hdu="BACKGROUND")
edisp = EnergyDispersion2D.read(irf_filename, hdu="ENERGY DISPERSION")

print(bkg)


######################################################################
# Note that the background is given in FoV coordinates with ``fov_lon``
# and ``fov_lat`` axis, and not in ``offset`` from the camera center. We
# can also check the Field of view alignment. Currently, two possible
# alignments are supported: alignment with the horizontal coordinate
# system (ALTAZ) and alignment with the equatorial coordinate system
# (RADEC).
#

print(bkg.fov_alignment)


######################################################################
# To evaluate the IRFs, pass the values for each axis. To know the default
# interpolation scheme for the data
#

print(bkg.interp_kwargs)

# Evaluate background
# Note that evaluate functions support  numpy-style broadcasting
energy = [1, 10, 100, 1000] * u.TeV
fov_lon = [1, 2] * u.deg
fov_lat = [1, 2] * u.deg
ev = bkg.evaluate(
    energy=energy.reshape(-1, 1, 1),
    fov_lat=fov_lat.reshape(1, -1, 1),
    fov_lon=fov_lon.reshape(1, 1, -1),
)
print(ev)


######################################################################
# We can customise the interpolation scheme. Here, we adapt to fill
# ``nan`` instead of ``0`` for extrapolated values
#

bkg.interp_kwargs["fill_value"] = np.nan

ev2 = bkg.evaluate(
    energy=energy.reshape(-1, 1, 1),
    fov_lat=fov_lat.reshape(1, -1, 1),
    fov_lon=fov_lon.reshape(1, 1, -1),
)
print(ev2)


######################################################################
# The interpolation scheme along each axis is taken from the `~gammapy.maps.MapAxis`
# specification. eg
#

print(
    "Interpolation scheme for energy axis is: ",
    bkg.axes["energy"].interp,
    "and for the fov_lon axis is: ",
    bkg.axes["fov_lon"].interp,
)

# Evaluate energy dispersion
ev = edisp.evaluate(energy_true=1 * u.TeV, offset=[0, 1] * u.deg, migra=[1, 1.2])
print(ev)

edisp.peek()

print(psf)


######################################################################
# The point spread function for the CTA DC1 is stored as a combination of 3
# Gaussians. Other PSFs, like a ``PSF_TABLE`` and analytic expressions
# like KING function are also supported. All PSF classes inherit from a
# common base ``PSF`` class.
#

print(psf.axes.names)

# To get the containment radius for a fixed fraction at a given position
print(
    psf.containment_radius(
        fraction=0.68, energy_true=1.0 * u.TeV, offset=[0.2, 4.0] * u.deg
    )
)

# Alternatively, to get the containment fraction for at a given position
print(
    psf.containment(
        rad=0.05 * u.deg, energy_true=1.0 * u.TeV, offset=[0.2, 4.0] * u.deg
    )
)


######################################################################
# Support for Asymmetric IRFs
# ---------------------------
#


######################################################################
# While Gammapy does not have inbuilt classes for supporting asymmetric
# IRFs (except for `~gammapy.irf.Background3D`), custom classes can be created. For
# this to work correctly with the `~gammapy.makers.MapDatasetMaker`, only variations
# with ``fov_lon`` and ``fov_lat`` can be allowed.
#
# The main idea is that the list of required axes should be correctly
# mentioned in the class definition.
#


######################################################################
# Effective Area
# --------------
#


class EffectiveArea3D(IRF):
    tag = "aeff_3d"
    required_axes = ["energy_true", "fov_lon", "fov_lat"]
    default_unit = u.m**2


energy_axis = MapAxis.from_energy_edges(
    [0.1, 0.3, 1.0, 3.0, 10.0] * u.TeV, name="energy_true"
)

nbin = 7
fov_lon_axis = MapAxis.from_edges([-1.5, -0.5, 0.5, 1.5] * u.deg, name="fov_lon")
fov_lat_axis = MapAxis.from_edges([-1.5, -0.5, 0.5, 1.5] * u.deg, name="fov_lat")

data = np.ones((4, 3, 3))
for i in range(1, 4):
    data[i] = data[i - 1] * 2.0
    data[i][-1] = data[i][0] * 1.2


aeff_3d = EffectiveArea3D(
    [energy_axis, fov_lon_axis, fov_lat_axis], data=data, unit=u.m**2
)
print(aeff_3d)

res = aeff_3d.evaluate(
    fov_lon=[-0.5, 0.8] * u.deg,
    fov_lat=[-0.5, 1.0] * u.deg,
    energy_true=[0.2, 8.0] * u.TeV,
)
print(res)

# to visualise at a given energy
# sphinx_gallery_thumbnail_number = 1
aeff_eval = aeff_3d.evaluate(energy_true=[1.0] * u.TeV)

ax = plt.subplot()
with quantity_support():
    caxes = ax.pcolormesh(
        fov_lat_axis.edges, fov_lon_axis.edges, aeff_eval.value.squeeze()
    )
fov_lat_axis.format_plot_xaxis(ax)
fov_lon_axis.format_plot_yaxis(ax)
ax.set_title("Asymmetric effective area")
plt.show()


######################################################################
# Unless specified, it is assumed these IRFs are in the RADEC frame
#

aeff_3d.fov_alignment


######################################################################
# Serialisation
# ~~~~~~~~~~~~~
#
# For serialisation, we need to add the class definition to the
# ``IRF_DL3_HDU_SPECIFICATION`` dictionary
#

IRF_DL3_HDU_SPECIFICATION["aeff_3d"] = {
    "extname": "EFFECTIVE AREA",
    "column_name": "EFFAREA",
    "mandatory_keywords": {
        **COMMON_IRF_HEADERS,
        "HDUCLAS2": "EFF_AREA",
        "HDUCLAS3": "FULL-ENCLOSURE",  # added here to have HDUCLASN in order
        "HDUCLAS4": "AEFF_3D",
    },
}

aeff_3d.write("test_aeff3d.fits", overwrite=True)

aeff_new = EffectiveArea3D.read("test_aeff3d.fits")
print(aeff_new)


######################################################################
# Create exposure map (DL4 product)
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# DL4 data products can be created from these IRFs.
#

axis = MapAxis.from_energy_bounds(0.1 * u.TeV, 10 * u.TeV, 6, name="energy_true")
pointing = SkyCoord(2, 1, unit="deg")
geom = WcsGeom.create(npix=(4, 3), binsz=2, axes=[axis], skydir=pointing)

print(geom)

exposure_map = make_map_exposure_true_energy(
    pointing=pointing, livetime="42 h", aeff=aeff_3d, geom=geom
)

exposure_map.plot_grid(add_cbar=True, figsize=(17, 7))
plt.show()


######################################################################
# Energy Dispersion
# -----------------
#


class EnergyDispersion3D(IRF):
    tag = "edisp_3d"
    required_axes = ["energy_true", "migra", "fov_lon", "fov_lat"]
    default_unit = u.one


######################################################################
# Note that most functions defined on the inbuilt IRF classes can be
# easily generalised to higher dimensions.
#

# Make a test case
energy_axis_true = MapAxis.from_energy_bounds(
    "0.1 TeV", "100 TeV", nbin=3, name="energy_true"
)

migra_axis = MapAxis.from_bounds(0, 4, nbin=2, node_type="edges", name="migra")

fov_lon_axis = MapAxis.from_edges([-1.5, -0.5, 0.5, 1.5] * u.deg, name="fov_lon")
fov_lat_axis = MapAxis.from_edges([-1.5, -0.5, 0.5, 1.5] * u.deg, name="fov_lat")

data = np.array(
    [
        [
            [
                [5.00e-01, 5.10e-01, 5.20e-01],
                [6.00e-01, 6.10e-01, 6.30e-01],
                [6.00e-01, 6.00e-01, 6.00e-01],
            ],
            [
                [2.0e-02, 2.0e-02, 2.0e-03],
                [2.0e-02, 2.0e-02, 2.0e-03],
                [2.0e-02, 2.0e-02, 2.0e-03],
            ],
        ],
        [
            [
                [5.00e-01, 5.10e-01, 5.20e-01],
                [6.00e-01, 6.10e-01, 6.30e-01],
                [6.00e-01, 6.00e-01, 6.00e-01],
            ],
            [
                [2.0e-02, 2.0e-02, 2.0e-03],
                [2.0e-02, 2.0e-02, 2.0e-03],
                [2.0e-02, 2.0e-02, 2.0e-03],
            ],
        ],
        [
            [
                [5.00e-01, 5.10e-01, 5.20e-01],
                [6.00e-01, 6.10e-01, 6.30e-01],
                [6.00e-01, 6.00e-01, 6.00e-01],
            ],
            [
                [3.0e-02, 6.0e-02, 2.0e-03],
                [3.0e-02, 5.0e-02, 2.0e-03],
                [3.0e-02, 4.0e-02, 2.0e-03],
            ],
        ],
    ]
)


edisp3d = EnergyDispersion3D(
    [energy_axis_true, migra_axis, fov_lon_axis, fov_lat_axis], data=data
)

print(edisp3d)

energy = [1, 2] * u.TeV
migra = np.array([0.98, 0.97, 0.7])
fov_lon = [0.1, 1.5] * u.deg
fov_lat = [0.0, 0.3] * u.deg

edisp_eval = edisp3d.evaluate(
    energy_true=energy.reshape(-1, 1, 1, 1),
    migra=migra.reshape(1, -1, 1, 1),
    fov_lon=fov_lon.reshape(1, 1, -1, 1),
    fov_lat=fov_lat.reshape(1, 1, 1, -1),
)
print(edisp_eval[0])


######################################################################
# Serialisation
# ~~~~~~~~~~~~~
#

IRF_DL3_HDU_SPECIFICATION["edisp_3d"] = {
    "extname": "ENERGY DISPERSION",
    "column_name": "MATRIX",
    "mandatory_keywords": {
        **COMMON_IRF_HEADERS,
        "HDUCLAS2": "EDISP",
        "HDUCLAS3": "FULL-ENCLOSURE",  # added here to have HDUCLASN in order
        "HDUCLAS4": "EDISP_3D",
    },
}

edisp3d.write("test_edisp.fits", overwrite=True)

edisp_new = EnergyDispersion3D.read("test_edisp.fits")
print(edisp_new)


######################################################################
# Create edisp kernel map (DL4 product) - `~gammapy.irf.EDispKernelMap`
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#

migra = MapAxis.from_edges(np.linspace(0.5, 1.5, 50), unit="", name="migra")
etrue = MapAxis.from_energy_bounds(0.5, 2, 6, unit="TeV", name="energy_true")
ereco = MapAxis.from_energy_bounds(0.5, 2, 3, unit="TeV", name="energy")
geom = WcsGeom.create(10, binsz=0.5, axes=[ereco, etrue], skydir=pointing)

edispmap = make_edisp_kernel_map(edisp3d, pointing, geom)

edispmap.peek()
plt.show()

#
print(edispmap.edisp_map.data[3][1][3])


######################################################################
# PSF
# ---
#
# There are two types of asymmetric PSFs that can be considered:
#
# -  asymmetry about the camera center: such PSF Tables can be supported,
# -  asymmetry about the source position: these PSF models cannot be supported correctly within
#    the data reduction scheme at present
#
# Also, analytic PSF models defined within the GADF scheme cannot be
# directly generalised to the 3D case for use within Gammapy.
#


class PSF_assym(IRF):
    tag = "psf_assym"
    required_axes = ["energy_true", "fov_lon", "fov_lat", "rad"]
    default_unit = u.sr**-1


energy_axis = MapAxis.from_energy_edges(
    [0.1, 0.3, 1.0, 3.0, 10.0] * u.TeV, name="energy_true"
)

nbin = 7
fov_lon_axis = MapAxis.from_edges([-1.5, -0.5, 0.5, 1.5] * u.deg, name="fov_lon")
fov_lat_axis = MapAxis.from_edges([-1.5, -0.5, 0.5, 1.5] * u.deg, name="fov_lat")

rad_axis = MapAxis.from_edges([0, 1, 2], unit="deg", name="rad")

data = 0.1 * np.ones((4, 3, 3, 2))
for i in range(1, 4):
    data[i] = data[i - 1] * 1.5


psf_assym = PSF_assym(
    axes=[energy_axis, fov_lon_axis, fov_lat_axis, rad_axis],
    data=data,
)
print(psf_assym)

energy = [1, 2] * u.TeV
rad = np.array([0.98, 0.97, 0.7]) * u.deg
fov_lon = [0.1, 1.5] * u.deg
fov_lat = [0.0, 0.3] * u.deg

psf_assym.evaluate(
    energy_true=energy.reshape(-1, 1, 1, 1),
    rad=rad.reshape(1, -1, 1, 1),
    fov_lon=fov_lon.reshape(1, 1, -1, 1),
    fov_lat=fov_lat.reshape(1, 1, 1, -1),
)


######################################################################
# Serialisation
# ~~~~~~~~~~~~~
#

IRF_DL3_HDU_SPECIFICATION["psf_assym"] = {
    "extname": "POINT SPREAD FUNCTION",
    "column_name": "MATRIX",
    "mandatory_keywords": {
        **COMMON_IRF_HEADERS,
        "HDUCLAS2": "PSF",
        "HDUCLAS3": "FULL-ENCLOSURE",
        "HDUCLAS4": "PSFnD",
    },
}

psf_assym.write("test_psf.fits.gz", overwrite=True)

psf_new = PSF_assym.read("test_psf.fits.gz")

######################################################################
# Create DL4 product - `~gammapy.irf.PSFMap`
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#

rad = MapAxis.from_edges(np.linspace(0.5, 3.0, 10), unit="deg", name="rad")
etrue = MapAxis.from_energy_bounds(0.5, 2, 6, unit="TeV", name="energy_true")
geom = WcsGeom.create(10, binsz=0.5, axes=[rad, etrue], skydir=pointing)

psfmap = make_psf_map(psf_assym, pointing, geom)

psfmap.peek()
plt.show()


######################################################################
# Containers for asymmetric analytic PSFs are not supported at present.
#
# **NOTE**: Support for asymmetric IRFs is preliminary at the moment, and
# will evolve depending on feedback.
#
#



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Source catalog and object base classes."""
import abc
import html
import numbers
from copy import deepcopy
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.utils import lazyproperty
from gammapy.maps import TimeMapAxis
from gammapy.modeling.models import Models
from gammapy.utils.table import table_row_to_dict

__all__ = ["SourceCatalog", "SourceCatalogObject"]


# https://pydanny.blogspot.com/2011/11/loving-bunch-class.html
class Bunch(dict):
    def __init__(self, **kw):
        dict.__init__(self, kw)
        self.__dict__.update(kw)


def format_flux_points_table(table):
    for column in table.colnames:
        if column.startswith(("dnde", "eflux", "flux", "e2dnde", "ref")):
            table[column].format = ".3e"
        elif column.startswith(
            ("e_min", "e_max", "e_ref", "sqrt_ts", "norm", "ts", "stat")
        ):
            table[column].format = ".3f"

    return table


class SourceCatalogObject:
    """Source catalog object.

    This class can be used directly, but it is mostly used as a
    base class for the other source catalog classes.

    The catalog data on this source is stored in the `source.data`
    attribute as a dict.

    The source catalog object is decoupled from the source catalog,
    it doesn't hold a reference back to it, except for a key
    ``_row_index`` of type ``int`` that links to the catalog table
    row the source information comes from.

    Parameters
    ----------
    data : dict
        Dictionary of data from a catalog for a given source.
    data_extended : dict
        Dictionary of data from a catalog for a given source in the case where the
        catalog contains an extended sources table (Fermi-LAT).
    data_spectral : dict
        Dictionary of data from a catalof for a given source in the case where the
        catalog contains a spectral table (Fermi-LAT 2PC and 3PC).
    """

    _source_name_key = "Source_Name"
    _row_index_key = "_row_index"

    def __init__(self, data, data_extended=None, data_spectral=None):
        self.data = Bunch(**data)
        if data_extended:
            self.data_extended = Bunch(**data_extended)
        self.data_spectral = Bunch(**data_spectral) if data_spectral else None

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    @property
    def name(self):
        """Source name as a string."""
        name = self.data[self._source_name_key]
        return name.strip()

    @property
    def row_index(self):
        """Row index of source in catalog as an integer."""
        return self.data[self._row_index_key]

    @property
    def position(self):
        """Source position as an `~astropy.coordinates.SkyCoord` object."""
        table = Table([self.data])
        return _skycoord_from_table(table)[0]


class SourceCatalog(abc.ABC):
    """Generic source catalog.

    This class can be used directly, but it is mostly used as a
    base class for the other source catalog classes.

    This is a thin wrapper around `~astropy.table.Table`,
    which is stored in the ``catalog.table`` attribute.

    Parameters
    ----------
    table : `~astropy.table.Table`
        Table with catalog data.
    source_name_key : str
        Column with source name information.
    source_name_alias : tuple of str
        Columns with source name aliases. This will allow accessing the source
        row by alias names as well.
    """

    @classmethod
    @abc.abstractmethod
    def description(cls):
        """Catalog description as a string."""
        pass

    @property
    @abc.abstractmethod
    def tag(self):
        pass

    source_object_class = SourceCatalogObject
    """Source class (`SourceCatalogObject`)."""

    def __init__(self, table, source_name_key="Source_Name", source_name_alias=()):
        self.table = table
        self._source_name_key = source_name_key
        self._source_name_alias = source_name_alias

    def __str__(self):
        return (
            f"{self.__class__.__name__}:\n"
            f"    name: {self.tag}\n"
            f"    description: {self.description}\n"
            f"    sources: {len(self.table)}\n"
        )

    @lazyproperty
    def _name_to_index_cache(self):
        # Make a dict for quick lookup: source name -> row index
        names = {}
        for idx, row in enumerate(self.table):
            name = row[self._source_name_key]
            names[name.strip()] = idx
            for alias_column in self._source_name_alias:
                for alias in str(row[alias_column]).split(","):
                    if not alias == "":
                        names[alias.strip()] = idx
        return names

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def row_index(self, name):
        """Look up row index of source by name.

        Parameters
        ----------
        name : str
            Source name.

        Returns
        -------
        index : int
            Row index of source in table.
        """
        index = self._name_to_index_cache[name]
        row = self.table[index]
        # check if name lookup is correct other wise recompute _name_to_index_cache

        possible_names = [row[self._source_name_key]]
        for alias_column in self._source_name_alias:
            possible_names += str(row[alias_column]).split(",")

        if name not in possible_names:
            self.__dict__.pop("_name_to_index_cache")
            index = self._name_to_index_cache[name]

        return index

    def source_name(self, index):
        """Look up source name by row index.

        Parameters
        ----------
        index : int
            Row index of source in table.
        """
        source_name_col = self.table[self._source_name_key]
        name = source_name_col[index]
        return name.strip()

    def __getitem__(self, key):
        """Get source by name.

        Parameters
        ----------
        key : str or int
            Source name or row index.

        Returns
        -------
        source : `SourceCatalogObject`
            An object representing one source.
        """
        if isinstance(key, str):
            index = self.row_index(key)
        elif isinstance(key, numbers.Integral):
            index = key
        elif isinstance(key, np.ndarray) and key.dtype == bool:
            new = deepcopy(self)
            new.table = self.table[key]
            return new
        else:
            raise TypeError(f"Invalid key: {key!r}, {type(key)}\n")

        return self._make_source_object(index)

    def _make_source_object(self, index):
        """Make one source object.

        Parameters
        ----------
        index : int
            Row index.

        Returns
        -------
        source : `SourceCatalogObject`
            Source object.
        """
        data = table_row_to_dict(self.table[index])
        data[SourceCatalogObject._row_index_key] = index

        fp_energy_edges = getattr(self, "flux_points_energy_edges", None)

        if fp_energy_edges is not None:
            data["fp_energy_edges"] = fp_energy_edges

        hist_table = getattr(self, "hist_table", None)
        hist2_table = getattr(self, "hist2_table", None)

        if hist_table:
            try:
                data["time_axis"] = TimeMapAxis.from_table(
                    hist_table, format="fermi-fgl"
                )
            except KeyError:
                pass

        if hist2_table:
            try:
                data["time_axis_2"] = TimeMapAxis.from_table(
                    hist2_table, format="fermi-fgl"
                )
            except KeyError:
                pass
        if "Extended_Source_Name" in data:
            name_extended = data["Extended_Source_Name"].strip()
        elif "Source_Name" in data:
            name_extended = data["Source_Name"].strip()
        else:
            name_extended = None

        name_spectral = self._get_name_spectral(data)

        try:
            idx = self._lookup_additional_table(
                self.extended_sources_table[self._source_name_key]
            )[name_extended]
            data_extended = table_row_to_dict(self.extended_sources_table[idx])
        except (KeyError, AttributeError):
            data_extended = None

        try:
            idx = self._lookup_additional_table(
                self.spectral_table[self._get_source_name_key]
            )[name_spectral]
            data_spectral = table_row_to_dict(self.spectral_table[idx])
        except (KeyError, AttributeError):
            data_spectral = None

        source = self.source_object_class(data, data_extended, data_spectral)
        return source

    @property
    def _get_source_name_key(self):
        return self._source_name_key

    def _get_name_spectral(self, data):
        if "Source_name" in data:
            name_spectral = data["Source_name"].strip()
        else:
            name_spectral = None
        return name_spectral

    def _lookup_additional_table(self, selected_table):
        """"""
        names = [_.strip() for _ in selected_table]
        idx = range(len(names))
        return dict(zip(names, idx))

    @property
    def positions(self):
        """Source positions as a `~astropy.coordinates.SkyCoord` object."""
        return _skycoord_from_table(self.table)

    def to_models(self, **kwargs):
        """Create Models object from catalog."""
        return Models([_.sky_model(**kwargs) for _ in self])


def _skycoord_from_table(table):
    keys = table.colnames

    if {"RAJ2000", "DEJ2000"}.issubset(keys):
        lon, lat, frame = "RAJ2000", "DEJ2000", "icrs"
    elif {"RAJ2000", "DECJ2000"}.issubset(keys):
        lon, lat, frame = "RAJ2000", "DECJ2000", "fk5"
    elif {"RA", "DEC"}.issubset(keys):
        lon, lat, frame = "RA", "DEC", "icrs"
    elif {"ra", "dec"}.issubset(keys):
        lon, lat, frame = "ra", "dec", "icrs"
    elif {"RAJD", "DECJD"}.issubset(keys):
        lon, lat, frame = "RAJD", "DECJD", "icrs"
    else:
        raise KeyError("No column GLON / GLAT or RA / DEC or RAJ2000 / DEJ2000 found.")

    unit = table[lon].unit.to_string() if table[lon].unit else "deg"

    return SkyCoord(table[lon], table[lat], unit=unit, frame=frame)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Other coordinate and distance-related functions."""
import numpy as np
from astropy.units import Quantity, Unit

__all__ = [
    "cartesian",
    "D_SUN_TO_GALACTIC_CENTER",
    "galactic",
    "motion_since_birth",
    "polar",
    "velocity_glon_glat",
]

# TODO: replace this with the default from the Galactocentric frame in astropy.coordinates
D_SUN_TO_GALACTIC_CENTER = Quantity(8.5, "kpc")
"""Default assumed distance from the Sun to the Galactic center (`~astropy.units.Quantity`)"""


def cartesian(r, theta):
    """Convert polar coordinates to cartesian coordinates."""
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y


def polar(x, y):
    """Convert cartesian coordinates to polar coordinates."""
    r = np.sqrt(x**2 + y**2)
    theta = np.arctan2(y, x)
    return r, theta


def galactic(x, y, z, obs_pos=None):
    """Compute galactic coordinates lon, lat and distance.

    For given position in cartesian coordinates (kpc).
    """
    obs_pos = obs_pos or [D_SUN_TO_GALACTIC_CENTER, 0, 0]
    y_prime = y + D_SUN_TO_GALACTIC_CENTER
    d = np.sqrt(x**2 + y_prime**2 + z**2)
    glon = np.arctan2(x, y_prime).to("deg")
    glat = np.arcsin(z / d).to("deg")
    return d, glon, glat


def velocity_glon_glat(x, y, z, vx, vy, vz):
    """
    Compute projected angular velocity in galactic coordinates.

    Parameters
    ----------
    x, y, z : `~astropy.units.Quantity`
        Position in x, y, z direction.
    vx, vy, vz : `~astropy.units.Quantity`
        Velocity in x, y, z direction.

    Returns
    -------
    v_glon, v_glat : `~astropy.units.Quantity`
        Projected velocity in Galactic sky coordinates.
    """
    y_prime = y + D_SUN_TO_GALACTIC_CENTER
    d = np.sqrt(x**2 + y_prime**2 + z**2)
    r = np.sqrt(x**2 + y_prime**2)

    v_glon = (-y_prime * vx + x * vy) / r**2
    v_glat = vz / (np.sqrt(1 - (z / d) ** 2) * d) - np.sqrt(
        vx**2 + vy**2 + vz**2
    ) * z / (np.sqrt(1 - (z / d) ** 2) * d**2)
    return v_glon * Unit("rad"), v_glat * Unit("rad")


def motion_since_birth(v, age, theta, phi):
    """
    Compute motion of an astrophysical object with a given velocity, direction and age.

    Parameters
    ----------
    v : `~astropy.units.Quantity`
        Absolute value of the velocity.
    age : `~astropy.units.Quantity`
        Age of the source.
    theta, phi : `~astropy.units.Quantity`
        Angular direction of the velocity.

    Returns
    -------
    dx, dy, dz : `~astropy.units.Quantity`
        Displacement in x, y, z direction.
    vx, vy, vz : `~astropy.units.Quantity`
        Velocity in x, y, z direction.
    """
    vx = v * np.cos(phi) * np.sin(theta)
    vy = v * np.sin(phi) * np.sin(theta)
    vz = v * np.cos(theta)

    # Compute new positions
    dx = vx * age
    dy = vy * age
    dz = vz * age
    return dx, dy, dz, vx, vy, vz



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from numpy.testing import assert_allclose
from astropy.coordinates import SkyCoord
from astropy.io import fits
from pydantic import ValidationError
from gammapy.data import EventListMetaData, ObservationMetaData, GTIMetaData
from gammapy.utils.metadata import (
    ObsInfoMetaData,
    PointingInfoMetaData,
    TargetMetaData,
    CreatorMetaData,
)
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import requires_data


@pytest.fixture()
def hess_eventlist_header():
    filename = make_path(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_023523.fits.gz"
    )
    hdulist = fits.open(filename)
    return hdulist["EVENTS"].header


def test_observation_metadata():
    obs_info = {
        "obs_id": 0,
        "telescope": "cta-north",
        "instrument": "lst",
        "observation_mode": "wobble",
    }
    target = {
        "name": "Crab",
        "position": SkyCoord(83.6287, 22.0147, unit="deg", frame="icrs"),
    }
    time_info = {
        "reference_time": "2023-01-01 00:00:00",
        "time_start": "2024-01-01 00:00:00",
        "time_stop": "2024-01-01 00:30:00",
    }

    input = {
        "obs_info": ObsInfoMetaData(**obs_info),
        "pointing": PointingInfoMetaData(),
        "location": "ctao_north",
        "deadtime_fraction": 0.05,
        "time_info": time_info,
        "target": TargetMetaData(**target),
        "optional": dict(test=0.5, other=True),
    }
    meta = ObservationMetaData(**input)

    assert meta.obs_info.telescope == "cta-north"
    assert meta.obs_info.instrument == "lst"
    assert meta.obs_info.observation_mode == "wobble"
    assert_allclose(meta.location.lon.value, -17.892005)
    assert meta.target.name == "Crab"
    assert_allclose(meta.target.position.ra.deg, 83.6287)
    assert_allclose(meta.time_info.time_stop.mjd, 60310.020833333)
    assert meta.optional["other"] is True

    with pytest.raises(ValidationError):
        meta.deadtime_fraction = 2.0

    with pytest.raises(ValidationError):
        meta.target.position = "J1749-2901"

    meta.target.position = None
    assert meta.target.position is None

    input_bad = input.copy()
    input_bad["location"] = "bad"

    with pytest.raises(ValueError):
        ObservationMetaData(**input_bad)


@requires_data()
def test_observation_metadata_from_header(hess_eventlist_header):
    meta = ObservationMetaData.from_header(hess_eventlist_header, format="gadf")

    assert meta.obs_info.telescope == "HESS"
    assert_allclose(meta.pointing.altaz_mean.alt.deg, 41.389789)
    assert_allclose(meta.time_info.time_start.mjd, 53343.92234)
    assert_allclose(meta.time_info.time_stop.mjd, 53343.941866)
    assert meta.target.name == "Crab Nebula"
    assert_allclose(meta.location.lat.deg, -23.271778)
    assert "TELLIST" in meta.optional
    assert meta.optional["TELLIST"] == "1,2,3,4"


@requires_data()
def test_observation_metadata_bad(hess_eventlist_header):
    # TODO: adapt with proper format handling
    with pytest.raises(ValueError):
        ObservationMetaData.from_header(hess_eventlist_header, format="bad")

    hess_eventlist_header.pop("DEADC")
    with pytest.raises(KeyError):
        ObservationMetaData.from_header(hess_eventlist_header, format="gadf")


def test_eventlist_metadata():
    input = {
        "event_class": "std",
        "optional": {"DST_VER": "v1.0", "ANA_VER": "v2.2", "CAL_VER": "v1.9"},
    }

    meta = EventListMetaData(**input)

    assert meta.event_class == "std"
    assert meta.optional["DST_VER"] == "v1.0"
    assert meta.optional["ANA_VER"] == "v2.2"
    assert meta.optional["CAL_VER"] == "v1.9"

    input_bad = input.copy()
    input_bad["location"] = "bad"
    with pytest.raises(ValueError):
        EventListMetaData(**input_bad)


def test_gti_metadata():
    input = {
        "reference_time": "2023-01-01 00:00:00",
        "creation": CreatorMetaData(creator="Test"),
    }
    meta = GTIMetaData(**input)

    hdr = meta.to_header()

    assert meta.creation.creator == "Test"
    assert_allclose(meta.reference_time.mjd, 59945.0)

    assert hdr["MJDREFI"] == 59945
    assert hdr["TIMESYS"] == "utc"
    assert hdr["CREATOR"] == "Test"

    meta_new = GTIMetaData.from_header(hdr)
    assert_allclose(meta_new.reference_time.mjd, 59945.0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.table import Column, Table
from astropy.time import Time
from gammapy.datasets import MapDataset
from gammapy.estimators import ExcessMapEstimator, FluxPoints
from gammapy.estimators.utils import (
    compute_lightcurve_doublingtime,
    compute_lightcurve_fpp,
    compute_lightcurve_fvar,
    compute_lightcurve_discrete_correlation,
    find_peaks,
    find_peaks_in_flux_map,
    get_rebinned_axis,
    resample_energy_edges,
)
from gammapy.maps import Map, MapAxis
from gammapy.utils.testing import assert_time_allclose, requires_data


@pytest.fixture()
def lc():
    meta = dict(TIMESYS="utc", SED_TYPE="flux")

    table = Table(
        meta=meta,
        data=[
            Column(Time(["2010-01-01", "2010-01-03"]).mjd, "time_min"),
            Column(Time(["2010-01-03", "2010-01-10"]).mjd, "time_max"),
            Column([[1.0, 2.0], [1.0, 2.0]], "e_min", unit="TeV"),
            Column([[2.0, 5.0], [2.0, 5.0]], "e_max", unit="TeV"),
            Column([[1e-11, 4e-12], [3e-11, 7e-12]], "flux", unit="cm-2 s-1"),
            Column(
                [[0.1e-11, 0.4e-12], [0.3e-11, 0.7e-12]], "flux_err", unit="cm-2 s-1"
            ),
            Column([[np.nan, np.nan], [3.6e-11, 1e-11]], "flux_ul", unit="cm-2 s-1"),
            Column([[False, False], [True, True]], "is_ul"),
            Column([[True, True], [True, True]], "success"),
        ],
    )

    return FluxPoints.from_table(table=table, format="lightcurve")


@pytest.fixture(scope="session")
def lc2():
    meta = dict(TIMESYS="utc", SED_TYPE="flux")

    table = Table(
        meta=meta,
        data=[
            Column(Time(["2010-01-01", "2010-01-03", "2010-01-05"]).mjd, "time_min"),
            Column(Time(["2010-01-03", "2010-01-05", "2010-01-07"]).mjd, "time_max"),
            Column([[1.0, 2.0], [1.0, 2.0], [1.0, 2.0]], "e_min", unit="GeV"),
            Column([[2.0, 5.0], [2.0, 5.0], [2.0, 5.0]], "e_max", unit="GeV"),
            Column(
                [[1.51e-7, 3.4e-8], [3.1e-7, 6.7e-8], [3.1e-7, 7.5e-8]],
                "flux",
                unit="m-2 s-1",
            ),
            Column(
                [[0.1e-7, 0.4e-8], [0.3e-7, 0.7e-8], [0.31e-7, 0.72e-8]],
                "flux_err",
                unit="m-2 s-1",
            ),
            Column(
                [[np.nan, np.nan], [3.6e-7, 1e-7], [3.6e-7, 1e-7]],
                "flux_ul",
                unit="m-2 s-1",
            ),
            Column([[False, False], [True, True], [True, True]], "is_ul"),
            Column([[True, True], [True, True], [True, True]], "success"),
        ],
    )

    return FluxPoints.from_table(table=table, format="lightcurve")


class TestFindPeaks:
    def test_simple(self):
        """Test a simple example"""
        image = Map.create(npix=(10, 5), unit="s")
        image.data[3, 3] = 11
        image.data[3, 4] = 10
        image.data[3, 5] = 12
        image.data[3, 6] = np.nan
        image.data[0, 9] = 1e20

        table = find_peaks(image, threshold=3)

        assert len(table) == 3
        assert table["value"].unit == "s"
        assert table["ra"].unit == "deg"
        assert table["dec"].unit == "deg"

        row = table[0]
        assert tuple((row["x"], row["y"])) == (9, 0)
        assert_allclose(row["value"], 1e20)
        assert_allclose(row["ra"], 359.55)
        assert_allclose(row["dec"], -0.2)

        row = table[1]
        assert tuple((row["x"], row["y"])) == (5, 3)
        assert_allclose(row["value"], 12)

    def test_no_peak(self):
        image = Map.create(npix=(10, 5))
        image.data[3, 5] = 12

        table = find_peaks(image, threshold=12.1)
        assert len(table) == 0

    def test_constant(self):
        image = Map.create(npix=(10, 5))

        table = find_peaks(image, threshold=3)
        assert len(table) == 0

    def test_flat_map(self):
        """Test a simple example"""
        axis1 = MapAxis.from_edges([1, 2], name="axis1")
        axis2 = MapAxis.from_edges([9, 10], name="axis2")
        image = Map.create(npix=(10, 5), unit="s", axes=[axis1, axis2])
        image.data[..., 3, 3] = 11
        image.data[..., 3, 4] = 10
        image.data[..., 3, 5] = 12
        image.data[..., 3, 6] = np.nan
        image.data[..., 0, 9] = 1e20

        table = find_peaks(image, threshold=3)
        row = table[0]

        assert len(table) == 3
        assert_allclose(row["value"], 1e20)
        assert_allclose(row["ra"], 359.55)
        assert_allclose(row["dec"], -0.2)


class TestFindFluxPeaks:
    """Tests for find_peaks_in_flux_map"""

    @requires_data()
    def test_find_peaks_in_flux_map(self):
        """Test a simple example"""
        dataset = MapDataset.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
        estimator = ExcessMapEstimator(
            correlation_radius="0.1 deg", energy_edges=[0.1, 10] * u.TeV
        )
        maps = estimator.run(dataset)
        table = find_peaks_in_flux_map(maps, threshold=5, min_distance=0.1 * u.deg)

        assert table["ra"].unit == "deg"
        assert table["dec"].unit == "deg"

    @requires_data()
    def test_no_peak(self):
        dataset = MapDataset.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
        position = SkyCoord(1.2, 1, frame="galactic", unit="deg")
        dataset_cutout = dataset.cutout(
            position=position, width=(0.2 * u.deg, 0.2 * u.deg)
        )

        estimator = ExcessMapEstimator(
            correlation_radius="0.1 deg", energy_edges=[0.1, 10] * u.TeV
        )
        maps = estimator.run(dataset_cutout)

        table = find_peaks_in_flux_map(maps, threshold=5, min_distance=0.1 * u.deg)

        assert len(table) == 0


def test_resample_energy_edges(spectrum_dataset):
    resampled_energy_edges = resample_energy_edges(spectrum_dataset, conditions={})
    assert (resampled_energy_edges == spectrum_dataset._geom.axes["energy"].edges).all()

    with pytest.raises(ValueError):
        resample_energy_edges(
            spectrum_dataset,
            conditions={"counts_min": spectrum_dataset.counts.data.sum() + 1},
        )

    resampled_energy_edges = resample_energy_edges(
        spectrum_dataset,
        conditions={"excess_min": spectrum_dataset.excess.data[-1] + 1},
    )
    grouped = spectrum_dataset.resample_energy_axis(
        MapAxis.from_edges(edges=resampled_energy_edges, name="energy")
    )

    assert grouped.counts.data.shape == (29, 1, 1)
    assert_allclose(np.squeeze(grouped.counts)[-1], 2518.0)
    assert_allclose(np.squeeze(grouped.background)[-1], 200)


def test_compute_lightcurve_fvar(lc):
    fvar = compute_lightcurve_fvar(lc)
    ffvar = fvar["fvar"].quantity
    ffvar_err = fvar["fvar_err"].quantity

    assert_allclose(ffvar, [[[0.698212]], [[0.37150576]]])
    assert_allclose(ffvar_err, [[[0.0795621]], [[0.074706]]])


def test_compute_lightcurve_fpp(lc):
    fpp = compute_lightcurve_fpp(lc)
    ffpp = fpp["fpp"].quantity
    ffpp_err = fpp["fpp_err"].quantity

    assert_allclose(ffpp, [[[0.99373035]], [[0.53551551]]])
    assert_allclose(ffpp_err, [[[0.07930673]], [[0.07397653]]])


def test_compute_lightcurve_doublingtime(lc):
    dtime = compute_lightcurve_doublingtime(lc)
    ddtime = dtime["doublingtime"].quantity
    ddtime_err = dtime["doubling_err"].quantity
    dcoord = dtime["doubling_coord"]

    assert_allclose(ddtime, [[[245305.49]], [[481572.59]]] * u.s)
    assert_allclose(ddtime_err, [[[45999.766]], [[11935.665]]] * u.s)
    assert_time_allclose(
        dcoord,
        Time([[[55197.99960648]], [[55197.99960648]]], format="mjd", scale="utc"),
    )


def test_compute_dcf(lc, lc2):
    dict = compute_lightcurve_discrete_correlation(lc, lc2, tau=3 * u.d)

    assert_allclose(dict["bins"], [-388800.0, -129600.0, 129600.0, 388800.0] * u.s)
    assert_allclose(
        dict["discrete_correlation"],
        [
            [-0.760599, -1.052783],
            [-0.760599, -0.537134],
            [1.014132, 1.059945],
            [-1.521198, -1.589918],
        ],
        rtol=1e-6,
    )
    assert_allclose(
        dict["discrete_correlation_err"],
        [[np.nan, np.nan], [np.nan, np.nan], [0.310513, 0.372241], [np.nan, np.nan]],
        rtol=1e-6,
    )

    dict2 = compute_lightcurve_discrete_correlation(lc2, tau=3 * u.d)
    assert_allclose(dict2["bins"], [-388800.0, -129600.0, 129600.0, 388800.0] * u.s)
    assert_allclose(
        dict2["discrete_correlation"],
        [
            [-1.11074, -1.448801],
            [-0.277685, -0.124862],
            [0.55537, 0.629465],
            [-1.11074, -1.448801],
        ],
        rtol=1e-5,
    )
    assert_allclose(
        dict2["discrete_correlation_err"],
        [[np.nan, np.nan], [1.178118, 0.868782], [0.589059, 0.53472], [np.nan, np.nan]],
        rtol=1e-6,
    )

    dict3 = compute_lightcurve_discrete_correlation(lc2)
    assert_allclose(dict3["bins"], [-345600.0, -115200.0, 115200.0, 345600.0] * u.s)


@requires_data()
def test_get_rebinned_axis():
    lc_1d = FluxPoints.read(
        "$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits",
        format="lightcurve",
    )
    axis_new = get_rebinned_axis(
        lc_1d, method="fixed-bins", group_size=2, axis_name="time"
    )
    assert_allclose(axis_new.bin_width[0], 20 * u.min)

    axis_new = get_rebinned_axis(
        lc_1d, method="min-ts", ts_threshold=2500.0, axis_name="time"
    )
    assert_allclose(axis_new.bin_width, [50, 30, 30, 50, 110, 70] * u.min)

    with pytest.raises(ValueError):
        get_rebinned_axis(lc_1d, method="error", value=2, axis_name="time")



"""Plot an energy dispersion using a gaussian parametrisation"""

import matplotlib.pyplot as plt
from gammapy.irf import EDispKernel
from gammapy.maps import MapAxis

energy_axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=10)
energy_axis_true = MapAxis.from_energy_bounds(
    "0.5 TeV", "30 TeV", nbin=10, per_decade=True, name="energy_true"
)


edisp = EDispKernel.from_gauss(
    energy_axis=energy_axis, energy_axis_true=energy_axis_true, sigma=0.1, bias=0
)
edisp.peek()
plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Metadata base container for Gammapy."""

import json
from typing import ClassVar, Literal, Optional, get_args
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.time import Time
import yaml
from pydantic import BaseModel, ConfigDict, Field, ValidationError, field_validator
from gammapy.utils.fits import skycoord_from_dict
from gammapy.utils.time import (
    time_ref_from_dict,
    time_ref_to_dict,
    time_relative_to_ref,
)
from gammapy.version import version
from .types import AltAzSkyCoordType, ICRSSkyCoordType, SkyCoordType, TimeType

__all__ = [
    "MetaData",
    "CreatorMetaData",
    "ObsInfoMetaData",
    "PointingInfoMetaData",
    "TimeInfoMetaData",
    "TargetMetaData",
]

METADATA_FITS_KEYS = {
    "creator": {
        "creator": "CREATOR",
        "date": {
            "input": lambda v: v.get("CREATED"),
            "output": lambda v: {"CREATED": v.iso},
        },
        "origin": "ORIGIN",
    },
    "obs_info": {
        "telescope": "TELESCOP",
        "instrument": "INSTRUME",
        "observation_mode": "OBS_MODE",
        "obs_id": "OBS_ID",
    },
    "pointing": {
        "radec_mean": {
            "input": lambda v: skycoord_from_dict(v, frame="icrs", ext="PNT"),
            "output": lambda v: {"RA_PNT": v.ra.deg, "DEC_PNT": v.dec.deg},
        },
        "altaz_mean": {
            "input": lambda v: skycoord_from_dict(v, frame="altaz", ext="PNT"),
            "output": lambda v: {"ALT_PNT": v.alt.deg, "AZ_PNT": v.az.deg},
        },
    },
    "target": {
        "name": "OBJECT",
        "position": {
            "input": lambda v: skycoord_from_dict(v, frame="icrs", ext="OBJ"),
            "output": lambda v: {"RA_OBJ": v.ra.deg, "DEC_OBJ": v.dec.deg},
        },
    },
}


class MetaData(BaseModel):
    """Base model for all metadata classes in Gammapy."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra="forbid",
        validate_default=True,
        use_enum_values=True,
    )

    @property
    def tag(self):
        """Returns MetaData tag."""
        return self._tag

    def to_header(self, format="gadf"):
        """Export MetaData to a FITS header.

        Conversion is performed following the definition in the METADATA_FITS_EXPORT_KEYS.

        Parameters
        ----------
        format : {'gadf'}
            Header format. Default is 'gadf'.

        Returns
        -------
        header : dict
            The header dictionary.
        """
        if format != "gadf":
            raise ValueError(f"Metadata to header: format {format} is not supported.")

        hdr_dict = {}

        fits_export_keys = METADATA_FITS_KEYS.get(self.tag)

        if fits_export_keys is None:
            # TODO: Should we raise an exception or simply a warning and return empty dict?
            raise TypeError(f"No FITS export is defined for metadata {self.tag}.")

        for key, item in fits_export_keys.items():
            value = self.model_dump().get(key)
            if value is not None:
                if not isinstance(item, str):
                    # Not a one to one conversion
                    hdr_dict.update(item["output"](value))
                else:
                    hdr_dict[item] = value

        extra_keys = set(self.__class__.model_fields.keys()) - set(
            fits_export_keys.keys()
        )

        for key in extra_keys:
            entry = getattr(self, key)
            if isinstance(entry, MetaData):
                hdr_dict.update(entry.to_header(format))
        return hdr_dict

    @classmethod
    def from_header(cls, header, format="gadf"):
        """Import MetaData from a FITS header.

        Conversion is performed following the definition in the METADATA_FITS_EXPORT_KEYS.

        Parameters
        ----------
        header : dict
            The header dictionary.
        format : {'gadf'}
            Header format. Default is 'gadf'.
        """
        # TODO: implement storage of optional metadata
        if format != "gadf":
            raise ValueError(f"Metadata from header: format {format} is not supported.")

        fits_export_keys = METADATA_FITS_KEYS.get(cls._tag)

        if fits_export_keys is None:
            raise TypeError(f"No FITS export is defined for metadata {cls._tag}.")

        kwargs = {}

        for key, item in fits_export_keys.items():
            if not isinstance(item, str):
                # Not a one to one conversion
                kwargs[key] = item["input"](header)
            else:
                kwargs[key] = header.get(item)

        extra_keys = set(cls.model_fields.keys()) - set(fits_export_keys.keys())

        for key in extra_keys:
            value = cls.model_fields[key]
            args = get_args(value.annotation)
            try:
                if issubclass(args[0], MetaData):
                    kwargs[key] = args[0].from_header(header, format)
            except TypeError:
                pass

        try:
            return cls(**kwargs)
        except ValidationError:
            return cls.model_construct(**kwargs)

    def to_yaml(self):
        """Dump metadata content to yaml."""
        meta = {"metadata": json.loads(self.model_dump_json())}
        return yaml.dump(
            meta, sort_keys=False, indent=4, width=80, default_flow_style=False
        )


class CreatorMetaData(MetaData):
    """Metadata containing information about the object creation.

    Parameters
    ----------
    creator : str
        The software used to create the data contained in the parent object.
        Default is the used Gammapy version.
    date : `~astropy.time.Time` or str
        The creation date. Default is the current date.
    origin : str
        The organization at the origin of the data.
    """

    _tag: ClassVar[Literal["creator"]] = "creator"
    creator: Optional[str] = f"Gammapy {version}"
    date: Optional[TimeType] = Field(default_factory=Time.now)
    origin: Optional[str] = None

    def update_time(self):
        """Change creation date to Time.now()."""
        self.date = Time.now()


class ObsInfoMetaData(MetaData):
    """General metadata information about the observation.

    Parameters
    ----------
    obs_id : str or int
        The observation identifier.
    telescope : str, optional
        The telescope/observatory name.
    instrument : str, optional
        The specific instrument used.
    sub_array : str, optional
        The specific sub-array used.
    observation_mode : str, optional
        The observation mode.
    """

    _tag: ClassVar[Literal["obs_info"]] = "obs_info"

    obs_id: int
    telescope: Optional[str] = None
    instrument: Optional[str] = None
    sub_array: Optional[str] = None
    observation_mode: Optional[str] = None


class PointingInfoMetaData(MetaData):
    """General metadata information about the pointing.

    Parameters
    ----------
    radec_mean : `~astropy.coordinates.SkyCoord`, optional
        Mean pointing position of the observation in `icrs` frame.
    altaz_mean : `~astropy.coordinates.SkyCoord`, or `~astropy.coordinates.AltAz`, optional
        Mean pointing position of the observation in local AltAz frame.
    """

    _tag: ClassVar[Literal["pointing"]] = "pointing"

    radec_mean: Optional[ICRSSkyCoordType] = None
    altaz_mean: Optional[AltAzSkyCoordType] = None


class TimeInfoMetaData(MetaData):
    """General metadata information about the time range of an observation.

    Parameters
    ----------
    time_start : `~astropy.time.Time` or str
        The observation start time.
    time_stop : `~astropy.time.Time` or str
        The observation stop time.
    reference_time : `~astropy.time.Time` or str
        The observation reference time.
    """

    _tag: ClassVar[Literal["time_info"]] = "time_info"

    time_start: Optional[TimeType] = None
    time_stop: Optional[TimeType] = None
    reference_time: Optional[TimeType] = None

    def to_header(self, format="gadf"):
        if self.reference_time is None:
            return {}
        result = time_ref_to_dict(self.reference_time)
        if self.time_start is not None:
            result["TSTART"] = time_relative_to_ref(self.time_start, result).to_value(
                "s"
            )
        if self.time_stop is not None:
            result["TSTOP"] = time_relative_to_ref(self.time_stop, result).to_value("s")
        return result

    @classmethod
    def from_header(cls, header, format="gadf"):
        kwargs = {}
        try:
            time_ref = time_ref_from_dict(header)
        except KeyError:
            return cls()

        kwargs["reference_time"] = time_ref
        if "TSTART" in header:
            kwargs["time_start"] = time_ref + header["TSTART"] * u.s
        if "TSTOP" in header:
            kwargs["time_stop"] = time_ref + header["TSTOP"] * u.s

        return cls(**kwargs)


class TargetMetaData(MetaData):
    """General metadata information about the target.

    Parameters
    ----------
    name : str, optional
        The target name.
    position : `~astropy.coordinates.SkyCoord`, optional
        Position of the observation in `icrs` frame.

    """

    _tag: ClassVar[Literal["target"]] = "target"
    name: Optional[str] = None
    position: Optional[SkyCoordType] = None

    @field_validator("position", mode="after")
    def validate_radec_mean(cls, v):
        if isinstance(v, SkyCoord):
            return v.icrs



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

# NOTE: The configuration for the package, including the name, version, and
# other information are set in the setup.cfg file.
import sys

# First provide helpful messages if contributors try and run legacy commands
# for tests or docs.

TEST_HELP = """
Note: running tests is no longer done using 'python setup.py test'. Instead
you will need to run:

    tox -e test

If you don't already have tox installed, you can install it with:

    pip install tox

If you only want to run part of the test suite, you can also use pytest
directly with::

    pip install -e .[test]
    pytest

For more information, see:

  http://docs.astropy.org/en/latest/development/testguide.html#running-tests
"""

if "test" in sys.argv:
    print(TEST_HELP)
    sys.exit(1)

DOCS_HELP = """
Note: building the documentation is no longer done using
'python setup.py build_docs'. Instead you will need to run:

    tox -e build_docs

If you don't already have tox installed, you can install it with:

    pip install tox

You can also build the documentation with Sphinx directly using::

    pip install -e .[docs]
    cd docs
    make html

For more information, see:

  http://docs.astropy.org/en/latest/install.html#builddocs
"""

if "build_docs" in sys.argv or "build_sphinx" in sys.argv:
    print(DOCS_HELP)
    sys.exit(1)


# imports here so that people get the nice error messages above without needing
# build dependencies
import numpy as np  # noqa: E402
from Cython.Build import cythonize  # noqa: E402
from setuptools import Extension, setup  # noqa: E402

kwargs = dict(
    include_dirs=[np.get_include()],
    define_macros=[
        # fixes a warning when compiling
        ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"),
        # defines the oldest numpy we want to be compatible with
        ("NPY_TARGET_VERSION", "NPY_1_21_API_VERSION"),
    ],
)

extensions = [
    Extension(
        "gammapy.stats.fit_statistics_cython",
        sources=["gammapy/stats/fit_statistics_cython.pyx"],
        **kwargs,
    ),
]

setup(
    ext_modules=cythonize(extensions),
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import html
import logging
import sys
import astropy.units as u
from astropy.coordinates import AltAz, Angle, EarthLocation, SkyCoord
from astropy.io import fits
from astropy.units import Quantity
from .scripts import make_path

log = logging.getLogger(__name__)

__all__ = ["earth_location_from_dict", "LazyFitsData", "HDULocation"]


class HDULocation:
    """HDU localisation, loading and Gammapy object mapper.

    This represents one row in `HDUIndexTable`.

    It's more a helper class, that is wrapped by `~gammapy.data.Observation`,
    usually those objects will be used to access data.

    See also `HDU index table <https://gamma-astro-data-formats.readthedocs.io/en/latest/data_storage/hdu_index/index.html#hdu-index>`__.
    """

    def __init__(
        self,
        hdu_class,
        base_dir=".",
        file_dir=None,
        file_name=None,
        hdu_name=None,
        cache=True,
        format=None,
    ):
        self.hdu_class = hdu_class
        self.base_dir = base_dir
        self.file_dir = file_dir
        self.file_name = file_name
        self.hdu_name = hdu_name
        self.cache = cache
        self.format = format

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def info(self, file=None):
        """Print some summary information to stdout."""
        if not file:
            file = sys.stdout
        print(f"HDU_CLASS = {self.hdu_class}", file=file)
        print(f"BASE_DIR = {self.base_dir}", file=file)
        print(f"FILE_DIR = {self.file_dir}", file=file)
        print(f"FILE_NAME = {self.file_name}", file=file)
        print(f"HDU_NAME = {self.hdu_name}", file=file)

    def path(self, abs_path=True):
        """Full filename path.

        Include ``base_dir`` if ``abs_path`` is True.
        """
        path = make_path(self.base_dir) / self.file_dir / self.file_name

        if abs_path and path.exists():
            return path
        else:
            return make_path(self.file_dir) / self.file_name

    def get_hdu(self):
        """Get HDU."""
        filename = self.path(abs_path=True)
        # Here we're intentionally not calling `with fits.open`
        # because we don't want the file to remain open.
        hdu_list = fits.open(str(filename), memmap=False)
        return hdu_list[self.hdu_name]

    def load(self):
        """Load HDU as appropriate class."""
        from gammapy.irf import IRF_REGISTRY

        hdu_class = self.hdu_class
        filename = self.path()
        hdu = self.hdu_name

        if hdu_class == "events":
            from gammapy.data import EventList

            return EventList.read(filename, hdu=hdu)
        elif hdu_class == "gti":
            from gammapy.data.gti import GTI

            return GTI.read(filename, hdu=hdu)
        elif hdu_class == "map":
            from gammapy.maps import Map

            return Map.read(filename, hdu=hdu, format=self.format)
        elif hdu_class == "pointing":
            # FIXME: support loading the pointing table
            from gammapy.data import FixedPointingInfo

            return FixedPointingInfo.read(filename, hdu=hdu)
        elif hdu_class == "observation_metadata":
            from gammapy.data import ObservationMetaData

            with fits.open(filename) as hdulist:
                header = hdulist[hdu].header
                return ObservationMetaData.from_header(header)
        else:
            cls = IRF_REGISTRY.get_cls(hdu_class)

            return cls.read(filename, hdu=hdu)


class LazyFitsData(object):
    """A lazy FITS data descriptor.

    Parameters
    ----------
    cache : bool
        Whether to cache the data.
    """

    def __init__(self, cache=True):
        self.cache = cache

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, objtype):
        if instance is None:
            # Accessed on a class, not an instance
            return self

        try:
            return instance.__dict__[self.name]
        except KeyError:
            hdu_loc = instance.__dict__[f"_{self.name}_hdu"]
            try:
                value = hdu_loc.load()
            except KeyError:
                value = None
                log.warning(f"HDU '{hdu_loc.hdu_name}' not found")
            if self.cache and hdu_loc.cache:
                instance.__dict__[self.name] = value
            return value

    def __set__(self, instance, value):
        if isinstance(value, HDULocation):
            instance.__dict__[f"_{self.name}_hdu"] = value
        else:
            instance.__dict__[self.name] = value


def earth_location_from_dict(meta):
    """Create `~astropy.coordinates.EarthLocation` from FITS header dictionary."""
    lon = Angle(meta["GEOLON"], "deg")
    lat = Angle(meta["GEOLAT"], "deg")
    if "GEOALT" in meta:
        height = Quantity(meta["GEOALT"], "meter")
    elif "ALTITUDE" in meta:
        height = Quantity(meta["ALTITUDE"], "meter")
    else:
        raise KeyError("The GEOALT or ALTITUDE header keyword must be set")

    return EarthLocation(lon=lon, lat=lat, height=height)


def earth_location_to_dict(location):
    """Convert `~astropy.coordinates.EarthLocation` to FITS header dictionary."""
    return {
        "GEOLON": location.lon.deg,
        "GEOLAT": location.lat.deg,
        "ALTITUDE": location.height.to_value(u.m),
    }


def skycoord_from_dict(header, frame="icrs", ext="PNT"):
    """Create `~astropy.coordinates.SkyCoord` from a dictionary of FITS keywords.

    Parameters
    ----------
    header : dict
        The input dictionary.
    frame : {"icrs", "galactic", "altaz"}
        The frame to use. Default is 'icrs'.
    ext: str, optional
        The keyword extension to apply to the keywords names. Default is 'PNT'.

    Returns
    -------
    skycoord : `~astropy.coordinates.skycoord`
        The input SkyCoord.
    """
    ext = "_" + ext if ext != "" else ""

    if frame == "altaz":
        alt = header.get("ALT" + ext, None)
        az = header.get("AZ" + ext, None)
        return (
            AltAz(alt=alt * u.deg, az=az * u.deg)
            if (alt is not None and az is not None)
            else None
        )
    elif frame == "icrs":
        coords = header.get("RA" + ext, None), header.get("DEC" + ext, None)
    elif frame == "galactic":
        coords = header.get("GLON" + ext, None), header.get("GLAT" + ext, None)
    else:
        raise ValueError(
            f"Unsupported frame {frame}. Select in 'icrs', 'galactic', 'altaz'."
        )
    if coords[0] is not None and coords[1] is not None:
        return SkyCoord(coords[0], coords[1], unit="deg", frame=frame)
    else:
        return None



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Dark matter spectra."""

import numpy as np
import astropy.units as u
from astropy.table import Table
from gammapy.maps import Map, MapAxis, RegionGeom
from gammapy.modeling import Parameter
from gammapy.modeling.models import SpectralModel, TemplateNDSpectralModel
from gammapy.utils.scripts import make_path

__all__ = ["PrimaryFlux", "DarkMatterAnnihilationSpectralModel"]


class PrimaryFlux(TemplateNDSpectralModel):
    """DM-annihilation gamma-ray spectra.

    Based on the precomputed models by Cirelli et al. (2016). All available
    annihilation channels can be found there. The dark matter mass will be set
    to the nearest available value. The spectra will be available as
    `~gammapy.modeling.models.TemplateNDSpectralModel` for a chosen dark matter mass and
    annihilation channel. Using a `~gammapy.modeling.models.TemplateNDSpectralModel`
    allows the interpolation between different dark matter masses.

    Parameters
    ----------
    mDM : `~astropy.units.Quantity`
        Dark matter particle mass as rest mass energy.
    channel: str
        Annihilation channel. List available channels with `~gammapy.spectrum.PrimaryFlux.allowed_channels`.

    References
    ----------
    * `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection"
      <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    * `Cirelli et al. (2016), "PPPC 4 DM ID: A Poor Particle Physicist Cookbook for Dark Matter Indirect Detection"
      <http://www.marcocirelli.net/PPPC4DMID.html>`_
    """

    channel_registry = {
        "eL": "eL",
        "eR": "eR",
        "e": "e",
        "muL": r"\[Mu]L",
        "muR": r"\[Mu]R",
        "mu": r"\[Mu]",
        "tauL": r"\[Tau]L",
        "tauR": r"\[Tau]R",
        "tau": r"\[Tau]",
        "q": "q",
        "c": "c",
        "b": "b",
        "t": "t",
        "WL": "WL",
        "WT": "WT",
        "W": "W",
        "ZL": "ZL",
        "ZT": "ZT",
        "Z": "Z",
        "g": "g",
        "gamma": r"\[Gamma]",
        "h": "h",
        "nu_e": r"\[Nu]e",
        "nu_mu": r"\[Nu]\[Mu]",
        "nu_tau": r"\[Nu]\[Tau]",
        "V->e": "V->e",
        "V->mu": r"V->\[Mu]",
        "V->tau": r"V->\[Tau]",
    }

    table_filename = "$GAMMAPY_DATA/dark_matter_spectra/AtProduction_gammas.dat"

    tag = ["PrimaryFlux", "dm-pf"]

    def __init__(self, mDM, channel):
        self.table_path = make_path(self.table_filename)
        if not self.table_path.exists():
            raise FileNotFoundError(
                f"\n\nFile not found: {self.table_filename}\n"
                "You may download the dataset needed with the following command:\n"
                "gammapy download datasets --src dark_matter_spectra"
            )
        else:
            self.table = Table.read(
                str(self.table_path),
                format="ascii.fast_basic",
                guess=False,
                delimiter=" ",
            )

        self.channel = channel

        # create RegionNDMap for channel

        masses = np.unique(self.table["mDM"])
        log10x = np.unique(self.table["Log[10,x]"])

        mass_axis = MapAxis.from_nodes(masses, name="mass", interp="log", unit="GeV")
        log10x_axis = MapAxis.from_nodes(log10x, name="energy_true")

        channel_name = self.channel_registry[self.channel]

        geom = RegionGeom(region=None, axes=[log10x_axis, mass_axis])
        region_map = Map.from_geom(
            geom=geom, data=self.table[channel_name].reshape(geom.data_shape)
        )

        interp_kwargs = {"extrapolate": True, "fill_value": 0, "values_scale": "lin"}
        super().__init__(region_map, interp_kwargs=interp_kwargs)
        self.mDM = mDM
        self.mass.frozen = True

    @property
    def mDM(self):
        """Dark matter mass."""
        return u.Quantity(self.mass.value, "GeV")

    @mDM.setter
    def mDM(self, mDM):
        unit = self.mass.unit
        _mDM = u.Quantity(mDM).to(unit)
        _mDM_val = _mDM.to_value(unit)

        min_mass = u.Quantity(self.mass.min, unit)
        max_mass = u.Quantity(self.mass.max, unit)

        if _mDM_val < self.mass.min or _mDM_val > self.mass.max:
            raise ValueError(
                f"The mass {_mDM} is out of the bounds of the model. Please choose a mass between {min_mass} < `mDM` < {max_mass}"
            )

        self.mass.value = _mDM_val

    @property
    def allowed_channels(self):
        """List of allowed annihilation channels."""
        return list(self.channel_registry.keys())

    @property
    def channel(self):
        """Annihilation channel as a string."""
        return self._channel

    @channel.setter
    def channel(self, channel):
        if channel not in self.allowed_channels:
            raise ValueError(
                f"Invalid channel: {channel}\nAvailable: {self.allowed_channels}\n"
            )
        else:
            self._channel = channel

    def evaluate(self, energy, **kwargs):
        """Evaluate the primary flux."""
        mass = {"mass": self.mDM}
        kwargs.update(mass)

        log10x = np.log10(energy / self.mDM)

        dN_dlogx = super().evaluate(log10x, **kwargs)
        dN_dE = dN_dlogx / (energy * np.log(10))
        return dN_dE


class DarkMatterAnnihilationSpectralModel(SpectralModel):
    r"""Dark matter annihilation spectral model.

    The gamma-ray flux is computed as follows:

    .. math::
        \frac{\mathrm d \phi}{\mathrm d E} =
        \frac{\langle \sigma\nu \rangle}{4\pi k m^2_{\mathrm{DM}}}
        \frac{\mathrm d N}{\mathrm dE} \times J(\Delta\Omega)

    Parameters
    ----------
    mass : `~astropy.units.Quantity`
        Dark matter mass.
    channel : str
        Annihilation channel for `~gammapy.astro.darkmatter.PrimaryFlux`, e.g. "b" for "bbar".
        See `PrimaryFlux.channel_registry` for more.
    scale : float
        Scale parameter for model fitting.
    jfactor : `~astropy.units.Quantity`
        Integrated J-Factor needed when `~gammapy.modeling.models.PointSpatialModel`
        is used.
    z: float
        Redshift value.
    k: int
        Type of dark matter particle (k:2 Majorana, k:4 Dirac).

    Examples
    --------
    This is how to instantiate a `DarkMatterAnnihilationSpectralModel` model::

        >>> import astropy.units as u
        >>> from gammapy.astro.darkmatter import DarkMatterAnnihilationSpectralModel

        >>> channel = "b"
        >>> massDM = 5000*u.Unit("GeV")
        >>> jfactor = 3.41e19 * u.Unit("GeV2 cm-5")
        >>> modelDM = DarkMatterAnnihilationSpectralModel(mass=massDM, channel=channel, jfactor=jfactor)  # noqa: E501

    References
    ----------
    `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection"
    <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    """

    THERMAL_RELIC_CROSS_SECTION = 3e-26 * u.Unit("cm3 s-1")
    """Thermally averaged annihilation cross-section"""

    scale = Parameter(
        "scale",
        1,
        unit="",
        interp="log",
    )
    tag = ["DarkMatterAnnihilationSpectralModel", "dm-annihilation"]

    def __init__(self, mass, channel, scale=scale.quantity, jfactor=1, z=0, k=2):
        self.k = k
        self.z = z
        self.mass = u.Quantity(mass)
        self.channel = channel
        self.jfactor = u.Quantity(jfactor)
        self.primary_flux = PrimaryFlux(mass, channel=self.channel)
        super().__init__(scale=scale)

    def evaluate(self, energy, scale):
        """Evaluate dark matter annihilation model."""
        flux = (
            scale
            * self.jfactor
            * self.THERMAL_RELIC_CROSS_SECTION
            * self.primary_flux(energy=energy * (1 + self.z))
            / self.k
            / self.mass
            / self.mass
            / (4 * np.pi)
        )
        return flux

    def to_dict(self, full_output=False):
        """Convert to dictionary."""
        data = super().to_dict(full_output=full_output)
        data["spectral"]["channel"] = self.channel
        data["spectral"]["mass"] = self.mass.to_string()
        data["spectral"]["jfactor"] = self.jfactor.to_string()
        data["spectral"]["z"] = self.z
        data["spectral"]["k"] = self.k
        return data

    @classmethod
    def from_dict(cls, data):
        """Create spectral model from a dictionary.

        Parameters
        ----------
        data : dict
            Dictionary with model data.

        Returns
        -------
        model : `DarkMatterAnnihilationSpectralModel`
            Dark matter annihilation spectral model.
        """
        data = data["spectral"]
        data.pop("type")
        parameters = data.pop("parameters")
        scale = [p["value"] for p in parameters if p["name"] == "scale"][0]
        return cls(scale=scale, **data)


class DarkMatterDecaySpectralModel(SpectralModel):
    r"""Dark matter decay spectral model.

    The gamma-ray flux is computed as follows:

    .. math::
        \frac{\mathrm d \phi}{\mathrm d E} =
        \frac{\Gamma}{4\pi m_{\mathrm{DM}}}
        \frac{\mathrm d N}{\mathrm dE} \times J(\Delta\Omega)

    Parameters
    ----------
    mass : `~astropy.units.Quantity`
        Dark matter mass.
    channel : str
        Annihilation channel for `~gammapy.astro.darkmatter.PrimaryFlux`, e.g. "b" for "bbar".
        See `PrimaryFlux.channel_registry` for more.
    scale : float
        Scale parameter for model fitting
    jfactor : `~astropy.units.Quantity`
        Integrated J-Factor needed when `~gammapy.modeling.models.PointSpatialModel`
        is used.
    z: float
        Redshift value.

    Examples
    --------
    This is how to instantiate a `DarkMatterAnnihilationSpectralModel` model::

        >>> import astropy.units as u
        >>> from gammapy.astro.darkmatter import DarkMatterDecaySpectralModel

        >>> channel = "b"
        >>> massDM = 5000*u.Unit("GeV")
        >>> jfactor = 3.41e19 * u.Unit("GeV cm-2")
        >>> modelDM = DarkMatterDecaySpectralModel(mass=massDM, channel=channel, jfactor=jfactor)  # noqa: E501

    References
    ----------
    `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection"
    <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    """

    LIFETIME_AGE_OF_UNIVERSE = 4.3e17 * u.Unit("s")
    """Use age of univserse as lifetime"""

    scale = Parameter(
        "scale",
        1,
        unit="",
        interp="log",
    )

    tag = ["DarkMatterDecaySpectralModel", "dm-decay"]

    def __init__(self, mass, channel, scale=scale.quantity, jfactor=1, z=0):
        self.z = z
        self.mass = u.Quantity(mass)
        self.channel = channel
        self.jfactor = u.Quantity(jfactor)
        self.primary_flux = PrimaryFlux(mass, channel=self.channel)
        super().__init__(scale=scale)

    def evaluate(self, energy, scale):
        """Evaluate dark matter decay model."""
        flux = (
            scale
            * self.jfactor
            * self.primary_flux(energy=energy * (1 + self.z))
            / self.LIFETIME_AGE_OF_UNIVERSE
            / self.mass
            / (4 * np.pi)
        )
        return flux

    def to_dict(self, full_output=False):
        data = super().to_dict(full_output=full_output)
        data["spectral"]["channel"] = self.channel
        data["spectral"]["mass"] = self.mass.to_string()
        data["spectral"]["jfactor"] = self.jfactor.to_string()
        data["spectral"]["z"] = self.z
        return data

    @classmethod
    def from_dict(cls, data):
        """Create spectral model from dictionary.

        Parameters
        ----------
        data : dict
            Dictionary with model data.

        Returns
        -------
        model : `DarkMatterDecaySpectralModel`
            Dark matter decay spectral model.
        """
        data = data["spectral"]
        data.pop("type")
        parameters = data.pop("parameters")
        scale = [p["value"] for p in parameters if p["name"] == "scale"][0]
        return cls(scale=scale, **data)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from gammapy.scripts.main import cli
from gammapy.utils.testing import requires_dependency, run_cli


def test_cli_tutorial_info():
    result = run_cli(cli, ["tutorial", "setup", "--help"])
    assert "Usage" in result.output


@requires_dependency("requests")
@requires_dependency("tqdm")
@pytest.mark.remote_data
def test_cli_tutorial_info_no_args(tmp_path):
    result = run_cli(cli, ["tutorial", "setup"])
    assert "System" in result.output
    assert "Gammapy package" in result.output
    assert "Other packages" in result.output
    assert "Gammapy environment variables" in result.output



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from gammapy.utils.registry import Registry
from .core import Dataset, Datasets
from .flux_points import FluxPointsDataset
from .io import OGIPDatasetReader, OGIPDatasetWriter, FermipyDatasetsReader
from .map import (
    MapDataset,
    MapDatasetOnOff,
    create_empty_map_dataset_from_irfs,
    create_map_dataset_from_observation,
    create_map_dataset_geoms,
)
from .metadata import MapDatasetMetaData
from .simulate import MapDatasetEventSampler, ObservationEventSampler
from .spectrum import SpectrumDataset, SpectrumDatasetOnOff
from .utils import apply_edisp, split_dataset

DATASET_REGISTRY = Registry(
    [
        MapDataset,
        MapDatasetOnOff,
        SpectrumDataset,
        SpectrumDatasetOnOff,
        FluxPointsDataset,
    ]
)

"""Registry of dataset classes in Gammapy."""

__all__ = [
    "create_empty_map_dataset_from_irfs",
    "create_map_dataset_from_observation",
    "create_map_dataset_geoms",
    "Dataset",
    "DATASET_REGISTRY",
    "Datasets",
    "FermipyDatasetsReader",
    "FluxPointsDataset",
    "MapDataset",
    "MapDatasetEventSampler",
    "MapDatasetOnOff",
    "ObservationEventSampler",
    "OGIPDatasetWriter",
    "OGIPDatasetReader",
    "SpectrumDataset",
    "SpectrumDatasetOnOff",
    "MapDatasetMetaData",
    "apply_edisp",
    "split_dataset",
]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Make test reference data files."""

from gammapy.catalog import CATALOG_REGISTRY

cat = CATALOG_REGISTRY.get_cls("3fgl")()
open("3fgl_J0000.1+6545.txt", "w").write(str(cat["3FGL J0000.1+6545"]))
open("3fgl_J0001.4+2120.txt", "w").write(str(cat["3FGL J0001.4+2120"]))
open("3fgl_J0023.4+0923.txt", "w").write(str(cat["3FGL J0023.4+0923"]))
open("3fgl_J0835.3-4510.txt", "w").write(str(cat["3FGL J0835.3-4510"]))

cat = CATALOG_REGISTRY.get_cls("4fgl")(
    "$GAMMAPY_DATA/catalogs/fermi/gll_psc_v20.fit.gz"
)
open("4fgl_J0000.3-7355.txt", "w").write(str(cat["4FGL J0000.3-7355"]))
open("4fgl_J0001.5+2113.txt", "w").write(str(cat["4FGL J0001.5+2113"]))
open("4fgl_J0002.8+6217.txt", "w").write(str(cat["4FGL J0002.8+6217"]))
open("4fgl_J1409.1-6121e.txt", "w").write(str(cat["4FGL J1409.1-6121e"]))

cat = CATALOG_REGISTRY.get_cls("4fgl")()
open("4fgl-dr4_J0534.5+2200.txt", "w").write(str(cat["4FGL J0534.5+2200"]))

cat = CATALOG_REGISTRY.get_cls("2fhl")()
open("2fhl_j1445.1-0329.txt", "w").write(str(cat["2FHL J1445.1-0329"]))
open("2fhl_j0822.6-4250e.txt", "w").write(str(cat["2FHL J0822.6-4250e"]))

cat = CATALOG_REGISTRY.get_cls("3fhl")()
open("3fhl_j2301.9+5855e.txt", "w").write(str(cat["3FHL J2301.9+5855e"]))

cat = CATALOG_REGISTRY.get_cls("2hwc")()
open("2hwc_j0534+220.txt", "w").write(str(cat["2HWC J0534+220"]))
open("2hwc_j0631+169.txt", "w").write(str(cat["2HWC J0631+169"]))

cat = CATALOG_REGISTRY.get_cls("hgps")()
open("hess_j1713-397.txt", "w").write(str(cat["HESS J1713-397"]))
open("hess_j1825-137.txt", "w").write(str(cat["HESS J1825-137"]))
open("hess_j1930+188.txt", "w").write(str(cat["HESS J1930+188"]))

cat = CATALOG_REGISTRY.get_cls("gamma-cat")()
open("gammacat_hess_j1813-178.txt", "w").write(str(cat["HESS J1813-178"]))
open("gammacat_hess_j1848-018.txt", "w").write(str(cat["HESS J1848-018"]))
open("gammacat_vela_x.txt", "w").write(str(cat["Vela X"]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import scipy.ndimage
from scipy import special
from scipy.interpolate import InterpolatedUnivariateSpline
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
from gammapy.datasets import SpectrumDataset, SpectrumDatasetOnOff
from gammapy.datasets.map import MapEvaluator
from gammapy.maps import Map, MapAxis, TimeMapAxis, WcsNDMap
from gammapy.modeling import Parameter
from gammapy.modeling.models import (
    ConstantFluxSpatialModel,
    PowerLawSpectralModel,
    SkyModel,
)
from gammapy.stats import (
    compute_flux_doubling,
    compute_fpp,
    compute_fvar,
    discrete_correlation,
)
from gammapy.stats.utils import ts_to_sigma
from .map.core import FluxMaps

__all__ = [
    "combine_flux_maps",
    "combine_significance_maps",
    "estimate_exposure_reco_energy",
    "find_peaks",
    "find_peaks_in_flux_map",
    "resample_energy_edges",
    "get_rebinned_axis",
    "get_combined_flux_maps",
    "get_combined_significance_maps",
    "compute_lightcurve_fvar",
    "compute_lightcurve_fpp",
    "compute_lightcurve_doublingtime",
    "compute_lightcurve_discrete_correlation",
]


def find_peaks(image, threshold, min_distance=1):
    """Find local peaks in an image.

    This is a very simple peak finder, that finds local peaks
    (i.e. maxima) in images above a given ``threshold`` within
    a given ``min_distance`` around each given pixel.

    If you get multiple spurious detections near a peak, usually
    it's best to smooth the image a bit, or to compute it using
    a different method in the first place to result in a smooth image.
    You can also increase the ``min_distance`` parameter.

    The output table contains one row per peak and the following columns:

    - ``x`` and ``y`` are the pixel coordinates (first pixel at zero).
    - ``ra`` and ``dec`` are the RA / DEC sky coordinates (ICRS frame).
    - ``value`` is the pixel value.

    It is sorted by peak value, starting with the highest value.

    If there are no pixel values above the threshold, an empty table is returned.

    There are more featureful peak finding and source detection methods
    e.g. in the ``photutils`` or ``scikit-image`` Python packages.

    Parameters
    ----------
    image : `~gammapy.maps.WcsNDMap`
        Image like Map.
    threshold : float or array-like
        The data value or pixel-wise data values to be used for the
        detection threshold.  A 2D ``threshold`` must have the same
        shape as the map ``data``.
    min_distance : int or `~astropy.units.Quantity`
        Minimum distance between peaks. An integer value is interpreted
        as pixels. Default is 1.

    Returns
    -------
    output : `~astropy.table.Table`
        Table with parameters of detected peaks.

    Examples
    --------
    >>> import astropy.units as u
    >>> from gammapy.datasets import MapDataset
    >>> from gammapy.estimators import ExcessMapEstimator
    >>> from gammapy.estimators.utils import find_peaks
    >>>
    >>> dataset = MapDataset.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
    >>> estimator = ExcessMapEstimator(
    ...     correlation_radius="0.1 deg", energy_edges=[0.1, 10] * u.TeV
    ... )
    >>> maps = estimator.run(dataset)
    >>> # Find the peaks which are above 5 sigma
    >>> sources = find_peaks(maps["sqrt_ts"], threshold=5, min_distance="0.25 deg")
    """
    if not isinstance(image, WcsNDMap):
        raise TypeError("find_peaks only supports WcsNDMap")

    if not image.geom.is_flat:
        raise ValueError(
            "find_peaks only supports flat Maps, with no spatial axes of length 1."
        )

    if isinstance(min_distance, (str, u.Quantity)):
        min_distance = np.mean(u.Quantity(min_distance) / image.geom.pixel_scales)
        min_distance = np.round(min_distance).to_value("")

    size = 2 * min_distance + 1

    # Remove non-finite values to avoid warnings or spurious detection
    data = image.sum_over_axes(keepdims=False).data
    data[~np.isfinite(data)] = np.nanmin(data)

    # Handle edge case of constant data; treat as no peak
    if np.all(data == data.flat[0]):
        return Table()

    # Run peak finder
    data_max = scipy.ndimage.maximum_filter(data, size=size, mode="constant")
    mask = (data == data_max) & (data > threshold)
    y, x = mask.nonzero()
    value = data[y, x]

    # Make and return results table

    if len(value) == 0:
        return Table()

    coord = SkyCoord.from_pixel(x, y, wcs=image.geom.wcs).icrs

    table = Table()
    table["value"] = value * image.unit
    table["x"] = x
    table["y"] = y
    table["ra"] = coord.ra
    table["dec"] = coord.dec

    table["ra"].format = ".5f"
    table["dec"].format = ".5f"
    table["value"].format = ".5g"

    table.sort("value")
    table.reverse()

    return table


def find_peaks_in_flux_map(maps, threshold, min_distance=1):
    """Find local test statistic peaks for a given Map.

    Utilises the `~gammapy.estimators.utils.find_peaks` function to find various parameters from FluxMaps.

    Parameters
    ----------
    maps : `~gammapy.estimators.FluxMaps`
        Input flux map object.
    threshold : float or array-like
        The test statistic data value or pixel-wise test statistic data values to be used for the
        detection threshold.  A 2D ``threshold`` must have the same.
        shape as the map ``data``.
    min_distance : int or `~astropy.units.Quantity`
        Minimum distance between peaks. An integer value is interpreted
        as pixels. Default is 1.

    Returns
    -------
    output : `~astropy.table.Table`
        Table with parameters of detected peaks.

    Examples
    --------
    >>> import astropy.units as u
    >>> from gammapy.datasets import MapDataset
    >>> from gammapy.estimators import ExcessMapEstimator
    >>> from gammapy.estimators.utils import find_peaks_in_flux_map
    >>>
    >>> dataset = MapDataset.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
    >>> estimator = ExcessMapEstimator(
    ...     correlation_radius="0.1 deg", energy_edges=[0.1, 10]*u.TeV
    ... )
    >>> maps = estimator.run(dataset)
    >>> # Find the peaks which are above 5 sigma
    >>> sources = find_peaks_in_flux_map(maps, threshold=5, min_distance=0.1*u.deg)

    """
    quantity_for_peaks = maps["sqrt_ts"]

    if not isinstance(maps, FluxMaps):
        raise TypeError(
            f"find_peaks_in_flux_map expects FluxMaps input. Got {type(maps)} instead."
        )

    if not quantity_for_peaks.geom.is_flat:
        raise ValueError(
            "find_peaks_in_flux_map only supports flat Maps, with energy axis of length 1."
        )

    table = find_peaks(quantity_for_peaks, threshold, min_distance)

    if len(table) == 0:
        return Table()

    x = np.array(table["x"])
    y = np.array(table["y"])

    table.remove_column("value")

    for name in maps.available_quantities:
        values = maps[name].quantity
        peaks = values[0, y, x]
        table[name] = peaks

    flux_data = maps["flux"].quantity
    table["flux"] = flux_data[0, y, x]
    flux_err_data = maps["flux_err"].quantity
    table["flux_err"] = flux_err_data[0, y, x]

    for column in table.colnames:
        if column.startswith(("flux", "flux_err")):
            table[column].format = ".3e"
        elif column.startswith(
            (
                "npred",
                "npred_excess",
                "counts",
                "sqrt_ts",
                "norm",
                "ts",
                "norm_err",
                "stat",
                "stat_null",
            )
        ):
            table[column].format = ".5f"

    table.reverse()

    return table


def estimate_exposure_reco_energy(dataset, spectral_model=None, normalize=True):
    """Estimate an exposure map in reconstructed energy.

    Parameters
    ----------
    dataset : `~gammapy.datasets.MapDataset` or `~gammapy.datasets.MapDatasetOnOff`
        The input dataset.
    spectral_model : `~gammapy.modeling.models.SpectralModel`, optional
        Assumed spectral shape. If None, a Power Law of index 2 is assumed. Default is None.
    normalize : bool
        Normalize the exposure to the total integrated flux of the spectral model.
        When not normalized it directly gives the predicted counts from the spectral
        model. Default is True.

    Returns
    -------
    exposure : `~gammapy.maps.Map`
        Exposure map in reconstructed energy.

    Examples
    --------
    .. testcode::

        from gammapy.datasets import MapDataset
        from gammapy.estimators.utils import estimate_exposure_reco_energy
        from gammapy.modeling.models import PowerLawSpectralModel
        dataset = MapDataset.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
        spectral_model = PowerLawSpectralModel()
        reco_exposure = estimate_exposure_reco_energy(
            dataset,
            spectral_model=spectral_model,
        )
        print(reco_exposure)

    .. testoutput::

        WcsNDMap

        geom  : WcsGeom
        axes  : ['lon', 'lat', 'energy']
        shape : (320, 240, 10)
        ndim  : 3
        unit  : cm2 s
        dtype : float64
    """
    if spectral_model is None:
        spectral_model = PowerLawSpectralModel()

    model = SkyModel(
        spatial_model=ConstantFluxSpatialModel(), spectral_model=spectral_model
    )

    energy_axis = dataset._geom.axes["energy"]

    if dataset.edisp is not None:
        edisp = dataset.edisp.get_edisp_kernel(position=None, energy_axis=energy_axis)
    else:
        edisp = None

    eval = MapEvaluator(model=model, exposure=dataset.exposure, edisp=edisp)
    reco_exposure = eval.compute_npred()

    if normalize:
        ref_flux = spectral_model.integral(
            energy_axis.edges[:-1], energy_axis.edges[1:]
        )
        reco_exposure = reco_exposure / ref_flux[:, np.newaxis, np.newaxis]

    return reco_exposure


def _satisfies_conditions(info_dict, conditions):
    satisfies = True
    for key in conditions.keys():
        satisfies &= info_dict[key.strip("_min")] > conditions[key]
    return satisfies


def resample_energy_edges(dataset, conditions={}):
    """Return energy edges that satisfy given condition on the per bin statistics.

    Parameters
    ----------
    dataset : `~gammapy.datasets.SpectrumDataset` or `~gammapy.datasets.SpectrumDatasetOnOff`
        The input dataset.
    conditions : dict
        Keyword arguments containing the per-bin conditions used to resample the axis.
        Available options are: 'counts_min', 'background_min', 'excess_min', 'sqrt_ts_min',
        'npred_min', 'npred_background_min', 'npred_signal_min'. Default is {}.

    Returns
    -------
    energy_edges : list of `~astropy.units.Quantity`
        Energy edges for the resampled energy axis.

    Examples
    --------
    >>> from gammapy.datasets import Datasets, SpectrumDatasetOnOff
    >>> from gammapy.estimators.utils import resample_energy_edges
    >>>
    >>> datasets = Datasets()
    >>>
    >>> for obs_id in [23523, 23526]:
    ...     dataset = SpectrumDatasetOnOff.read(
    ...         f"$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs{obs_id}.fits"
    ...     )
    ...     datasets.append(dataset)
    >>>
    >>> spectrum_dataset = Datasets(datasets).stack_reduce()
    >>> # Resample the energy edges so the minimum sqrt_ts is 2
    >>> resampled_energy_edges = resample_energy_edges(
    ...     spectrum_dataset,
    ...     conditions={"sqrt_ts_min": 2}
    ... )
    """
    if not isinstance(dataset, (SpectrumDataset, SpectrumDatasetOnOff)):
        raise NotImplementedError(
            "This method is currently supported for spectral datasets only."
        )

    available_conditions = [
        "counts_min",
        "background_min",
        "excess_min",
        "sqrt_ts_min",
        "npred_min",
        "npred_background_min",
        "npred_signal_min",
    ]
    for key in conditions.keys():
        if key not in available_conditions:
            raise ValueError(
                f"Unrecognized option {key}. The available methods are: {available_conditions}."
            )

    axis = dataset.counts.geom.axes["energy"]
    energy_min_all, energy_max_all = dataset.energy_range_total
    energy_edges = [energy_max_all]

    while energy_edges[-1] > energy_min_all:
        for energy_min in reversed(axis.edges_min):
            if energy_min >= energy_edges[-1]:
                continue
            elif len(energy_edges) == 1 and energy_min == energy_min_all:
                raise ValueError("The given conditions cannot be met.")

            sliced = dataset.slice_by_energy(
                energy_min=energy_min, energy_max=energy_edges[-1]
            )

            with np.errstate(invalid="ignore"):
                info = sliced.info_dict()

            if _satisfies_conditions(info, conditions):
                energy_edges.append(energy_min)
                break
    return u.Quantity(energy_edges[::-1])


def compute_lightcurve_fvar(lightcurve, flux_quantity="flux"):
    """
    Compute the fractional excess variance of the input lightcurve.

    Internally calls the `~gammapy.stats.compute_fvar` function.


    Parameters
    ----------
    lightcurve : `~gammapy.estimators.FluxPoints`
        The lightcurve object.
    flux_quantity : str, optional
        Flux quantity to use for calculation. Should be 'dnde', 'flux', 'e2dnde' or 'eflux'. Default is 'flux'.

    Returns
    -------
    fvar : `~astropy.table.Table`
        Table of fractional excess variance and associated error for each energy bin of the lightcurve.

    Examples
    --------
    .. testcode::

        from gammapy.estimators import FluxPoints
        from gammapy.estimators.utils import compute_lightcurve_fvar

        lightcurve = FluxPoints.read(
            "$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits",
            format="lightcurve",
        )
        fvar = compute_lightcurve_fvar(lightcurve)
    """
    flux = getattr(lightcurve, flux_quantity)
    flux_err = getattr(lightcurve, flux_quantity + "_err")

    time_id = flux.geom.axes.index_data("time")

    fvar, fvar_err = compute_fvar(flux.data, flux_err.data, axis=time_id)

    significance = fvar / fvar_err

    energies = lightcurve.geom.axes["energy"].edges
    table = Table(
        [energies[:-1], energies[1:], fvar, fvar_err, significance],
        names=("min_energy", "max_energy", "fvar", "fvar_err", "significance"),
        meta=lightcurve.meta,
    )

    return table


def compute_lightcurve_fpp(lightcurve, flux_quantity="flux"):
    """
    Compute the point-to-point excess variance of the input lightcurve.

    Internally calls the `~gammapy.stats.compute_fpp` function

    Parameters
    ----------
    lightcurve : `~gammapy.estimators.FluxPoints`
        The lightcurve object.
    flux_quantity : str, optional
        Flux quantity to use for calculation. Should be 'dnde', 'flux', 'e2dnde' or 'eflux'. Default is 'flux'.

    Returns
    -------
    table : `~astropy.table.Table`
        Table of point-to-point excess variance and associated error for each energy bin of the lightcurve.

    Examples
    --------
    .. testcode::

        from gammapy.estimators import FluxPoints
        from gammapy.estimators.utils import compute_lightcurve_fpp

        lightcurve = FluxPoints.read(
            "$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits",
            format="lightcurve",
        )
        fpp = compute_lightcurve_fpp(lightcurve)
    """
    flux = getattr(lightcurve, flux_quantity)
    flux_err = getattr(lightcurve, flux_quantity + "_err")

    time_id = flux.geom.axes.index_data("time")

    fpp, fpp_err = compute_fpp(flux.data, flux_err.data, axis=time_id)

    significance = fpp / fpp_err

    energies = lightcurve.geom.axes["energy"].edges
    table = Table(
        [energies[:-1], energies[1:], fpp, fpp_err, significance],
        names=("min_energy", "max_energy", "fpp", "fpp_err", "significance"),
        meta=dict(quantity=flux_quantity),
    )

    return table


def compute_lightcurve_doublingtime(lightcurve, flux_quantity="flux"):
    """
    Compute the minimum characteristic flux doubling and halving time for the input lightcurve.

    Internally calls the `~gammapy.stats.compute_flux_doubling` function.

    The characteristic doubling time  is estimated to obtain the
    minimum variability timescale for the light curves in which
    rapid variations are clearly evident: for example it is useful in AGN flaring episodes.

    This quantity, especially for AGN flares, is often expressed
    as the pair of doubling time and halving time, or the minimum characteristic time
    for the rising and falling components respectively.

    Parameters
    ----------
    lightcurve : `~gammapy.estimators.FluxPoints`
        The lightcurve object.
    flux_quantity : str, optional
        Flux quantity to use for calculation. Should be 'dnde', 'flux', 'e2dnde' or 'eflux'.
        Default is 'flux'.

    Returns
    -------
    table : `~astropy.table.Table`
        Table of flux doubling/halving and associated error for each energy bin of the lightcurve
        with axis coordinates at which they were found.

    Examples
    --------
    .. testcode::

        from gammapy.estimators import FluxPoints
        from gammapy.estimators.utils import compute_lightcurve_doublingtime

        lightcurve = FluxPoints.read(
            "$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits",
            format="lightcurve",
        )
        doublingtime = compute_lightcurve_doublingtime(lightcurve)

    References
    ----------
    `Brown et al. (2013), "Locating the γ-ray emission region of the flat spectrum radio quasar PKS 1510−089"
    <https://ui.adsabs.harvard.edu/abs/2013MNRAS.431..824B/abstract>`_
    """
    flux = getattr(lightcurve, flux_quantity)
    flux_err = getattr(lightcurve, flux_quantity + "_err")
    coords = lightcurve.geom.axes["time"].center

    axis = flux.geom.axes.index_data("time")

    doubling_dict = compute_flux_doubling(flux.data, flux_err.data, coords, axis=axis)

    energies = lightcurve.geom.axes["energy"].edges
    table = Table(
        [
            energies[:-1],
            energies[1:],
            doubling_dict["doubling"],
            doubling_dict["doubling_err"],
            lightcurve.geom.axes["time"].reference_time
            + doubling_dict["doubling_coord"],
            doubling_dict["halving"],
            doubling_dict["halving_err"],
            lightcurve.geom.axes["time"].reference_time
            + doubling_dict["halving_coord"],
        ],
        names=(
            "min_energy",
            "max_energy",
            "doublingtime",
            "doubling_err",
            "doubling_coord",
            "halvingtime",
            "halving_err",
            "halving_coord",
        ),
        meta=dict(flux_quantity=flux_quantity),
    )

    return table


def compute_lightcurve_discrete_correlation(
    lightcurve1, lightcurve2=None, flux_quantity="flux", tau=None
):
    """Compute the discrete correlation function for two lightcurves, or the discrete autocorrelation if only one lightcurve is provided.

    NaN values will be ignored in the computation in order to account for possible gaps in the data.

    Internally calls the `~gammapy.stats.discrete_correlation` function.

    Parameters
    ----------
    lightcurve1 : `~gammapy.estimators.FluxPoints`
        The first lightcurve object.
    lightcurve2 : `~gammapy.estimators.FluxPoints`, optional
        The second lightcurve object. If not provided, the autocorrelation for the first lightcurve will be computed.
        Default is None.
    flux_quantity : str, optional
        Flux quantity to use for calculation. Should be 'dnde', 'flux', 'e2dnde' or 'eflux'.
        The choice does not affect the computation. Default is 'flux'.
    tau : `~astropy.units.Quantity`, optional
        Size of the bins to compute the discrete correlation.
        If None, the bin size will be double the bins of the first lightcurve. Default is None.

    Returns
    -------
    discrete_correlation_dict : dict
        Dictionary containing the discrete correlation results. Entries are:

                * "bins" : the array of discrete time bins
                * "discrete_correlation" : discrete correlation function values
                * "discrete_correlation_err" : associated error

    Examples
    --------
    .. testcode::

        from gammapy.estimators import FluxPoints
        from gammapy.estimators.utils import compute_lightcurve_discrete_correlation
        import astropy.units as u

        lightcurve = FluxPoints.read(
            "$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits",
            format="lightcurve",
        )
        discrete_correlation = compute_lightcurve_discrete_correlation(
            lightcurve,
            tau=3*u.d,
        )

    References
    ----------
    `Edelson et al. (1988), "THE DISCRETE CORRELATION FUNCTION: A NEW METHOD FOR ANALYZING
    UNEVENLY SAMPLED VARIABILITY DATA" <https://ui.adsabs.harvard.edu/abs/1988ApJ...333..646E/abstract>`_
    """
    flux1 = getattr(lightcurve1, flux_quantity)
    flux_err1 = getattr(lightcurve1, flux_quantity + "_err")
    coords1 = lightcurve1.geom.axes["time"].center
    axis = flux1.geom.axes.index_data("time")

    if tau is None:
        tau = (coords1[-1] - coords1[0]) / (0.5 * len(coords1))

    if lightcurve2:
        flux2 = getattr(lightcurve2, flux_quantity)
        flux_err2 = getattr(lightcurve2, flux_quantity + "_err")
        coords2 = lightcurve2.geom.axes["time"].center
        bins, dcf, dcf_err = discrete_correlation(
            flux1.data,
            flux_err1.data,
            flux2.data,
            flux_err2.data,
            coords1,
            coords2,
            tau,
            axis,
        )

    else:
        bins, dcf, dcf_err = discrete_correlation(
            flux1.data,
            flux_err1.data,
            flux1.data,
            flux_err1.data,
            coords1,
            coords1,
            tau,
            axis,
        )

    discrete_correlation_dict = {
        "bins": bins,
        "discrete_correlation": dcf,
        "discrete_correlation_err": dcf_err,
    }

    return discrete_correlation_dict


def get_edges_fixed_bins(fluxpoint, group_size, axis_name="energy"):
    """Rebin the flux point to combine value adjacent bins.

    Parameters
    ----------
    fluxpoint : `~gammapy.estimators.FluxPoints`
        The flux points object to rebin.
    group_size : int
        Number of bins to combine.
    axis_name : str, optional
        The axis name to combine along. Default is 'energy'.

    Returns
    -------
    edges_min : `~astropy.units.Quantity` or `~astropy.time.Time`
        Minimum bin edge for the new axis.
    edges_max : `~astropy.units.Quantity` or `~astropy.time.Time`
        Maximum bin edge for the new axis.
    """
    ax = fluxpoint.geom.axes[axis_name]
    nbin = ax.nbin
    if not isinstance(group_size, int):
        raise ValueError("Only integer number of bins can be combined")
    idx = np.arange(0, nbin, group_size)
    if idx[-1] < nbin:
        idx = np.append(idx, nbin)
    edges_min = ax.edges_min[idx[:-1]]
    edges_max = ax.edges_max[idx[1:] - 1]
    return edges_min, edges_max


def get_edges_min_ts(fluxpoint, ts_threshold, axis_name="energy"):
    """Rebin the flux point to combine adjacent bins until a minimum TS is obtained.

    Note that to convert TS to significance, it is necessary to take the number
    of degrees of freedom into account.


    Parameters
    ----------
    fluxpoint : `~gammapy.estimators.FluxPoints`
        The flux points object to rebin.
    ts_threshold : float
        The minimum significance desired.
    axis_name : str, optional
        The axis name to combine along. Default is 'energy'.

    Returns
    -------
    edges_min : `~astropy.units.Quantity` or `~astropy.time.Time`
        Minimum bin edge for the new axis.
    edges_max : `~astropy.units.Quantity` or `~astropy.time.Time`
        Maximum bin edge for the new axis.
    """
    ax = fluxpoint.geom.axes[axis_name]
    nbin = ax.nbin

    e_min, e_max = ax.edges_min[0], ax.edges_max[0]
    edges_min = np.zeros(nbin) * e_min.unit
    edges_max = np.zeros(nbin) * e_max.unit
    i, i1 = 0, 0
    while e_max < ax.edges_max[-1]:
        ts = fluxpoint.ts.data[i]
        e_min = ax.edges_min[i]
        while ts < ts_threshold and i < ax.nbin - 1:
            i = i + 1
            ts = ts + fluxpoint.ts.data[i]
        e_max = ax.edges_max[i]
        i = i + 1
        edges_min[i1] = e_min
        edges_max[i1] = e_max
        i1 = i1 + 1
    edges_max = edges_max[:i1]
    edges_min = edges_min[:i1]

    return edges_min, edges_max


RESAMPLE_METHODS = {
    "fixed-bins": get_edges_fixed_bins,
    "min-ts": get_edges_min_ts,
}


def get_rebinned_axis(fluxpoint, axis_name="energy", method=None, **kwargs):
    """Get the rebinned axis for resampling the flux point object along the mentioned axis.

    Parameters
    ----------
    fluxpoint : `~gammapy.estimators.FluxPoints`
        The flux point object to rebin.
    axis_name : str, optional
        The axis name to combine along. Default is 'energy'.
    method : str
        The method to resample the axis. Supported options are
        'fixed_bins' and 'min-ts'.
    kwargs : dict
        Keywords passed to `get_edges_fixed_bins` or `get_edges_min_ts`.
        If method is 'fixed-bins', keyword should be `group_size`.
        If method is 'min-ts', keyword should be `ts_threshold`.

    Returns
    -------
    axis_new : `~gammapy.maps.MapAxis` or `~gammapy.maps.TimeMapAxis`
        The new axis.

    Examples
    --------
    >>> from gammapy.estimators.utils import get_rebinned_axis
    >>> from gammapy.estimators import FluxPoints
    >>>
    >>> # Rebin lightcurve axis
    >>> lc_1d = FluxPoints.read(
    ...         "$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits",
    ...         format="lightcurve",
    ...     )
    >>> # Rebin axis by combining adjacent bins as per the group_size
    >>> new_axis = get_rebinned_axis(
    ...     lc_1d, method="fixed-bins", group_size=2, axis_name="time"
    ... )
    >>>
    >>> # Rebin HESS flux points axis
    >>> fp = FluxPoints.read(
    ...         "$GAMMAPY_DATA/estimators/crab_hess_fp/crab_hess_fp.fits"
    ... )
    >>> # Rebin according to a minimum significance
    >>> axis_new = get_rebinned_axis(
    ...     fp, method='min-ts', ts_threshold=4, axis_name='energy'
    ... )
    """
    # TODO: Make fixed_bins and fixed_edges work for multidimensions
    if not fluxpoint.geom.axes.is_unidimensional:
        raise ValueError(
            "Rebinning is supported only for Unidimensional FluxPoints \n "
            "Please use `iter_by_axis` to create Unidimensional FluxPoints"
        )

    if method not in RESAMPLE_METHODS.keys():
        raise ValueError("Incorrect option. Choose from", RESAMPLE_METHODS.keys())

    edges_min, edges_max = RESAMPLE_METHODS[method](
        fluxpoint=fluxpoint, axis_name=axis_name, **kwargs
    )
    ax = fluxpoint.geom.axes[axis_name]

    if isinstance(ax, TimeMapAxis):
        axis_new = TimeMapAxis.from_time_edges(
            time_min=edges_min + ax.reference_time,
            time_max=edges_max + ax.reference_time,
        )
    else:
        edges = np.append(edges_min, edges_max[-1])
        axis_new = MapAxis.from_edges(edges, name=axis_name, interp=ax.interp)
    return axis_new


def combine_significance_maps(maps):
    """Compute excess and significance for a set of datasets.

    The significance computation assumes that the model contains
    one degree of freedom per valid energy bin in each dataset.
    The method implemented here is valid under the assumption
    that the TS in each independent bin follows a Chi2 distribution,
    then the sum of the TS also follows a Chi2 distribution (with the sum of the degrees of freedom).

    See, Zhen (2014): https://www.sciencedirect.com/science/article/abs/pii/S0167947313003204,
    Lancaster (1961): https://onlinelibrary.wiley.com/doi/10.1111/j.1467-842X.1961.tb00058.x


    Parameters
    ----------
    maps : list of `~gammapy.estimators.FluxMaps`
        List of maps with the same geometry.

    Returns
    -------
    results : dict
        Dictionary with entries:

                * "significance" : joint significance map.
                * "df" : degree of freedom map (one norm per valid bin).
                * "npred_excess" : summed excess map.
                * "estimator_results" : dictionary containing the flux maps computed for each dataset.

    See Also
    --------
    get_combined_significance_maps : same method but computing the significance maps from estimators and datasets.

    """
    geom = maps[0].ts.geom.to_image()
    ts_sum = Map.from_geom(geom)
    ts_sum_sign = Map.from_geom(geom)
    npred_excess_sum = Map.from_geom(geom)
    df = Map.from_geom(geom)

    for result in maps:
        df += np.sum(result["ts"].data > 0, axis=0)  # one dof (norm) per valid bin
        ts_sum += result["ts"].reduce_over_axes()
        ts_sum_sign += (
            result["ts"] * np.sign(result["npred_excess"])
        ).reduce_over_axes()
        npred_excess_sum += result["npred_excess"].reduce_over_axes()

    significance = Map.from_geom(geom)
    significance.data = ts_to_sigma(ts_sum.data, df.data) * np.sign(ts_sum_sign)
    return dict(
        significance=significance,
        df=df,
        ts=ts_sum,
        npred_excess=npred_excess_sum,
        estimator_results=maps,
    )


def get_combined_significance_maps(estimator, datasets):
    """Compute excess and significance for a set of datasets.

    The significance computation assumes that the model contains
    one degree of freedom per valid energy bin in each dataset.
    This method implemented here is valid under the assumption
    that the TS in each independent bin follows a Chi2 distribution,
    then the sum of the TS also follows a Chi2 distribution (with the sum of degree of freedom).

    Parameters
    ----------
    estimator : `~gammapy.estimators.ExcessMapEstimator` or `~gammapy.estimators.TSMapEstimator`
        Excess Map Estimator or TS Map Estimator
    datasets : `~gammapy.datasets.Datasets`
        Datasets containing only `~gammapy.datasets.MapDataset`.

    Returns
    -------
    results : dict
        Dictionary with entries:

                * "significance" : joint significance map.
                * "df" : degree of freedom map (one norm per valid bin).
                * "npred_excess" : summed excess map.
                * "estimator_results" : dictionary containing the flux maps computed for each dataset.

    Examples
    --------
    >>> from gammapy.datasets import Datasets
    >>> from gammapy.estimators import ExcessMapEstimator
    >>> from gammapy.estimators.utils import get_combined_significance_maps
    >>> datasets = Datasets.read("$GAMMAPY_DATA/hawc/DL4/HAWC_pass4_public_Crab.yaml")
    >>> estimator = ExcessMapEstimator(correlation_radius="0.2 deg")
    >>> combined = get_combined_significance_maps(estimator, datasets)

    References
    ----------
    * `Chen & Nadarajah (2014), "On the optimally weighted z-test for combining probabilities from independent
      studies" <https://www.sciencedirect.com/science/article/abs/pii/S0167947313003204>`_
    * `Lancaster (1961), "The Combination of Probabilities: an Application of Orthonormal Functions"
      <https://onlinelibrary.wiley.com/doi/10.1111/j.1467-842X.1961.tb00058.x>`_


    See Also
    --------
    combine_significance_maps : same method but using directly the significance maps from estimators
    """
    from .map.excess import ExcessMapEstimator
    from .map.ts import TSMapEstimator

    if not isinstance(estimator, (ExcessMapEstimator, TSMapEstimator)):
        raise TypeError(
            f"estimator type should be ExcessMapEstimator or TSMapEstimator), got {type(estimator)} instead."
        )

    results = []
    for dataset in datasets:
        results.append(estimator.run(dataset))
    return combine_significance_maps(results)


def combine_flux_maps(
    maps, method="gaussian_errors", reference_model=None, dnde_scan_axis=None
):
    """Create a FluxMaps by combining a list of flux maps with the same geometry.

     This assumes the flux maps are independent measurements of the same true value.
     The GTI is stacked in the process.

    Parameters
    ----------
    maps : list of `~gammapy.estimators.FluxMaps`
        List of maps with the same geometry.
    method : str
        * gaussian_errors :
            Under the gaussian error approximation the likelihood is given by the gaussian distibution.
            The product of gaussians is also a gaussian so can derive dnde, dnde_err, and ts.
        * distrib :
            Likelihood profile approximation assuming that probabilities distributions for
            flux points correspond to asymmetric gaussians and for upper limits to complementary error functions.
            Use available quantities among dnde, dnde_err, dnde_errp, dnde_errn, dnde_ul, and ts.
        * profile :
            Sum the likelihood profile maps.
            The flux maps must contains the `stat_scan` maps.

        Default is "gaussian_errors" which is the faster but least accurate solution,
        "distrib"  will be more accurate if dnde_errp and dnde_errn are available,
        "profile"  will be even more accurate if "stat_scan" is available.

    reference_model : `~gammapy.modeling.models.SkyModel`, optional
        Reference model to use for conversions.
        Default is None and is will use the reference_model of the first FluxMaps in the list.

    dnde_scan_axis : `~gammapy.maps.MapAxis`
        Map axis providing the dnde values used to compute the profile.
        Default is None and it will be derived from the first FluxMaps in the list.
        Used only if `method` is distrib or profile.

    Returns
    -------
    flux_maps : `~gammapy.estimators.FluxMaps`
        Joint flux map.


    See Also
    --------
    get_combined_flux_maps : same method but using directly the flux maps from estimators

    """
    gtis = [map_.gti for map_ in maps if map_.gti is not None]
    if np.any(gtis):
        gti = gtis[0].copy()
        for k in range(1, len(gtis)):
            gti.stack(gtis[k])
    else:
        gti = None
    # TODO : change this once we have stackable metadata objets
    metas = [map_.meta for map_ in maps if map_.meta is not None]
    meta = {}
    if np.any(metas):
        for data in metas:
            meta.update(data)
    if reference_model is None:
        reference_model = maps[0].reference_model

    if method == "gaussian_errors":
        means = [map_.dnde.copy() for map_ in maps]
        sigmas = [map_.dnde_err.copy() for map_ in maps]
        # compensate for the ts deviation from gaussian approximation expectation in each map
        ts_diff = np.nansum(
            [
                map_.ts.data - (map_.dnde.data / map_.dnde_err.data) ** 2
                for map_ in maps
            ],
            axis=0,
        )

        mean = means[0]
        sigma = sigmas[0]

        for k in range(1, len(means)):
            mean_k = means[k].quantity.to_value(mean.unit)
            sigma_k = sigmas[k].quantity.to_value(sigma.unit)

            mask_valid = np.isfinite(mean) & np.isfinite(sigma) & (sigma.data != 0)
            mask_valid_k = np.isfinite(mean_k) & np.isfinite(sigma_k) & (sigma_k != 0)
            mask = mask_valid & mask_valid_k
            mask_k = ~mask_valid & mask_valid_k

            mean.data[mask] = (
                (mean.data * sigma_k**2 + mean_k * sigma.data**2)
                / (sigma.data**2 + sigma_k**2)
            )[mask]
            sigma.data[mask] = (
                sigma.data * sigma_k / np.sqrt(sigma.data**2 + sigma_k**2)
            )[mask]

            mean.data[mask_k] = mean_k[mask_k]
            sigma.data[mask_k] = sigma_k[mask_k]

        ts = mean * mean / sigma / sigma + ts_diff
        ts.data[~np.isfinite(ts.data)] = np.nan

        kwargs = dict(
            sed_type="dnde", reference_model=reference_model, meta=meta, gti=gti
        )
        return FluxMaps.from_maps(dict(dnde=mean, dnde_err=sigma, ts=ts), **kwargs)

    elif method in ["distrib", "profile"]:
        if dnde_scan_axis is None:
            dnde_scan_axis = _default_scan_map(maps[0]).geom.axes["dnde"]
        for k, map_ in enumerate(maps):
            if method == "profile":
                map_stat_scan = interpolate_profile_map(map_, dnde_scan_axis)
            else:
                map_stat_scan = approximate_profile_map(map_, dnde_scan_axis)
            map_stat_scan.data[np.isnan(map_stat_scan.data)] = 0.0
            if k == 0:
                stat_scan = map_stat_scan
            else:
                stat_scan.data += map_stat_scan.data

        return get_flux_map_from_profile(
            {"stat_scan": stat_scan},
            reference_model=reference_model,
            meta=meta,
            gti=gti,
        )

    else:
        raise ValueError(
            f'Invalid method provided : {method}. Available methods are : "gaussian_errors", "distrib", "profile"'
        )


def get_combined_flux_maps(
    estimator,
    datasets,
    method="gaussian_errors",
    reference_model=None,
    dnde_scan_axis=None,
):
    """Create a `~gammapy.estimators.FluxMaps` by combining a list of flux maps with the same geometry.

    This assumes the flux maps are independent measurements of the same true value.
    The GTI is stacked in the process.

    Parameters
    ----------
    estimator : `~gammapy.estimators.ExcessMapEstimator` or `~gammapy.estimators.TSMapEstimator`
        Excess Map Estimator or TS Map Estimator
    datasets : `~gammapy.datasets.Datasets` or list of `~gammapy.datasets.MapDataset`
        Datasets containing only `~gammapy.datasets.MapDataset`.
    method : str
        * gaussian_errors :
            Under the gaussian error approximation the likelihood is given by the gaussian distibution.
            The product of gaussians is also a gaussian so can derive dnde, dnde_err, and ts.
        * distrib :
            Likelihood profile approximation assuming that probabilities distributions for
            flux points correspond to asymmetric gaussians and for upper limits to complementary error functions.
            Use available quantities among dnde, dnde_err, dnde_errp, dnde_errn, dnde_ul, and ts.
        * profile :
            Sum the likelihood profile maps.
            The flux maps must contain the `stat_scan` maps.

        Default is "gaussian_errors" which is the faster but least accurate solution,
        "distrib"  will be more accurate if dnde_errp and dnde_errn are available,
        "profile"  will be even more accurate if "stat_scan" is available.

    reference_model : `~gammapy.modeling.models.SkyModel`, optional
        Reference model to use for conversions.
        Default is None and is will use the reference_model of the first `~gammapy.estimators.FluxMaps` in the list.
    dnde_scan_axis : `~gammapy.maps.MapAxis`, optional
        Map axis providing the dnde values used to compute the profile.
        If None, it will be derived from the first FluxMaps in the list. Default is None.
        Used only if `method` is "distrib" or "profile".

    Returns
    -------
    results : dict
        Dictionary with entries:

                * "flux_maps" : `gammapy.estimators.FluxMaps`
                * "estimator_results" : dictionary containing the flux maps computed for each dataset.

    Examples
    --------
    >>> from gammapy.datasets import Datasets
    >>> from gammapy.estimators import TSMapEstimator
    >>> from gammapy.estimators.utils import get_combined_flux_maps
    >>> datasets = Datasets.read("$GAMMAPY_DATA/hawc/DL4/HAWC_pass4_public_Crab.yaml")
    >>> estimator = TSMapEstimator()
    >>> combined = get_combined_flux_maps(estimator, datasets)

    See Also
    --------
    combine_flux_maps : same method but using directly the flux maps from estimators

    """
    from .map.excess import ExcessMapEstimator
    from .map.ts import TSMapEstimator

    if not isinstance(estimator, (ExcessMapEstimator, TSMapEstimator)):
        raise TypeError(
            f"`estimator` type should be ExcessMapEstimator or TSMapEstimator), got {type(estimator)} instead."
        )

    results = []
    for dataset in datasets:
        results.append(estimator.run(dataset))

    output = dict()
    output["flux_maps"] = combine_flux_maps(
        results,
        method=method,
        reference_model=reference_model,
        dnde_scan_axis=dnde_scan_axis,
    )
    output["estimator_results"] = results
    return output


def _default_scan_map(flux_map, dnde_scan_axis=None):
    if dnde_scan_axis is None:
        dnde_scan_axis = MapAxis(
            _generate_scan_values() * flux_map.dnde_ref.squeeze(),
            interp="lin",
            node_type="center",
            name="dnde",
            unit=flux_map.dnde_ref.unit,
        )

    geom = flux_map.dnde.geom
    geom_scan = geom.to_image().to_cube([dnde_scan_axis] + list(geom.axes))
    return Map.from_geom(geom_scan, data=np.nan, unit="")


def interpolate_profile_map(flux_map, dnde_scan_axis=None):
    """Interpolate sparse likelihood profile to regular grid.

    Parameters
    ----------
    flux_map : `~gammapy.estimators.FluxMaps`
        Flux map.
    dnde_scan_axis : `~gammapy.maps.MapAxis`
        Map axis providing the dnde values used to compute the profile.
        Default is None and it will be derived from the flux_map.

    Returns
    -------
    scan_map: `~gammapy.estimators.Maps`
        Likelihood profile map.

    """
    stat_scan = _default_scan_map(flux_map, dnde_scan_axis)
    dnde_scan_axis = stat_scan.geom.axes["dnde"]

    mask_valid = ~np.isnan(flux_map.dnde.data)
    dnde_scan_values = flux_map.dnde_scan_values.quantity.to_value(dnde_scan_axis.unit)

    for ij, il, ik in zip(*np.where(mask_valid)):
        spline = InterpolatedUnivariateSpline(
            dnde_scan_values[ij, :, il, ik],
            flux_map.stat_scan.data[ij, :, il, ik],
            k=1,
            ext="raise",
            check_finite=True,
        )
        stat_scan.data[ij, :, il, ik] = spline(dnde_scan_axis.center)
    return stat_scan


def approximate_profile_map(
    flux_map, dnde_scan_axis=None, sqrt_ts_threshold_ul="ignore"
):
    """Likelihood profile approximation assuming that probabilities distributions for
    flux points correspond to asymmetric gaussians and for upper limits to complementary error functions.
    Use available quantities among dnde, dnde_err, dnde_errp, dnde_errn, dnde_ul and ts.

    Parameters
    ----------
    flux_map : `~gammapy.estimators.FluxMaps`
        Flux map.
    dnde_scan_axis : `~gammapy.maps.MapAxis`
        Map axis providing the dnde values used to compute the profile.
        Default is None and it will be derived from the flux_map.
    sqrt_ts_threshold_ul : int
        Threshold value in sqrt(TS) for upper limits.
        Default is `ignore` and no threshold is applied.
        Setting to `None` will use the one of `flux_map`.

    Returns
    -------
    scan_map: `~gammapy.estimators.Maps`
        Likelihood profile map.
    """
    stat_approx = _default_scan_map(flux_map, dnde_scan_axis)
    dnde_coord = stat_approx.geom.get_coord()["dnde"].value

    if sqrt_ts_threshold_ul is None:
        sqrt_ts_threshold_ul = flux_map.sqrt_ts_threshold_ul

    mask_valid = ~np.isnan(flux_map.dnde.data)
    ij, il, ik = np.where(mask_valid)
    loc = flux_map.dnde.data[mask_valid][:, None]
    value = dnde_coord[ij, :, il, ik]
    try:
        mask_p = dnde_coord >= flux_map.dnde.data
        mask_p2d = mask_p[ij, :, il, ik]
        new_axis = np.ones(mask_p2d.shape[1], dtype=bool)[None, :]
        scale = np.zeros(mask_p2d.shape)
        scale[mask_p2d] = (flux_map.dnde_errp.data[mask_valid][:, None] * new_axis)[
            mask_p2d
        ]
        scale[~mask_p2d] = (flux_map.dnde_errn.data[mask_valid][:, None] * new_axis)[
            ~mask_p2d
        ]
    except AttributeError:
        scale = flux_map.dnde_err.data[mask_valid]
        scale = scale[:, None]
    stat_approx.data[ij, :, il, ik] = ((value - loc) / scale) ** 2

    try:
        invalid_value = 999
        stat_min_p = (stat_approx.data + invalid_value * (~mask_p)).min(
            axis=1, keepdims=True
        )
        stat_min_m = (stat_approx.data + invalid_value * mask_p).min(
            axis=1, keepdims=True
        )

        mask_minp = mask_p & (stat_min_p > stat_min_m)
        stat_approx.data[mask_minp] = (stat_approx.data + stat_min_m - stat_min_p)[
            mask_minp
        ]
        mask_minn = ~mask_p & (stat_min_m >= stat_min_p)
        stat_approx.data[mask_minn] = (stat_approx.data + stat_min_p - stat_min_m)[
            mask_minn
        ]
    except NameError:
        pass

    if not sqrt_ts_threshold_ul == "ignore" and sqrt_ts_threshold_ul is not None:
        mask_ul = (flux_map.sqrt_ts.data < sqrt_ts_threshold_ul) & ~np.isnan(
            flux_map.dnde_ul.data
        )
        ij, il, ik = np.where(mask_ul)
        value = dnde_coord[ij, :, il, ik]
        loc_ul = flux_map.dnde_ul.data[mask_ul][:, None]
        scale_ul = flux_map.dnde_ul.data[mask_ul][:, None]
        stat_approx.data[ij, :, il, ik] = -2 * np.log(
            (special.erfc((-loc_ul + value) / scale_ul) / 2)
            / (special.erfc((-loc_ul + 0) / scale_ul) / 2)
        )

    stat_approx.data[np.isnan(stat_approx.data)] = np.inf
    stat_approx.data += -flux_map.ts.data - stat_approx.data.min(axis=1)
    return stat_approx


def get_flux_map_from_profile(
    flux_map, n_sigma=1, n_sigma_ul=2, reference_model=None, meta=None, gti=None
):
    """Create a new flux map using the likehood profile (stat_scan)
    to get ts, dnde, dnde_err, dnde_errp, dnde_errn, and dnde_ul.

    Parameters
    ----------
    flux_maps : `~gammapy.estimators.FluxMaps` or dict of `~gammapy.maps.WcsNDMap`
        Flux map or dict containing  a `stat_scan` entry
    n_sigma : float, optional
        Number of sigma for flux error. Must be a positive value. Default is 1.
    n_sigma_ul : float, optional
        Number of sigma for flux upper limits. Must be a positive value. Default is 2.
    reference_model : `~gammapy.modeling.models.SkyModel`, optional
        The reference model to use for conversions. If None, a model consisting
        of a point source with a power law spectrum of index 2 is assumed.
        Default is None and the one of `flux_map` will be used if available
    meta : dict, optional
        Dict of metadata.
        Default is None and the one of `flux_map` will be used if available
    gti : `~gammapy.data.GTI`, optional
        Maps GTI information.
        Default is None and the one of `flux_map` will be used if available

    Returns
    -------
    flux_maps : `~gammapy.estimators.FluxMaps`
        Flux map.
    """
    if isinstance(flux_map, dict):
        output_maps = flux_map
    else:
        if reference_model is None:
            reference_model = flux_map.reference_model
        if gti is None:
            gti = flux_map.gti
        if meta is None:
            meta = flux_map.meta

        output_maps = dict(
            stat_scan=flux_map.stat_scan, dnde_scan_values=flux_map.dnde_scan_values
        )

    if getattr(flux_map, "dnde_scan_values", False):
        dnde_coord = flux_map["dnde_scan_values"].quantity
    else:
        dnde_coord = flux_map["stat_scan"].geom.get_coord()["dnde"]

    geom = (
        flux_map["stat_scan"]
        .geom.to_image()
        .to_cube([flux_map["stat_scan"].geom.axes["energy"]])
    )

    ts = -flux_map["stat_scan"].data.min(axis=1) * u.Unit("")
    ind = flux_map["stat_scan"].data.argmin(axis=1)
    ij, ik, il = np.indices(ind.shape)
    dnde = dnde_coord[ij, ind, ik, il]

    maskp = dnde_coord > dnde
    stat_diff = flux_map["stat_scan"].data - flux_map["stat_scan"].data.min(axis=1)

    invalid_value = 999
    ind = np.abs(stat_diff + invalid_value * maskp - n_sigma**2).argmin(axis=1)
    dnde_errn = dnde - dnde_coord[ij, ind, ik, il]

    ind = np.abs(stat_diff + invalid_value * (~maskp) - n_sigma**2).argmin(axis=1)
    dnde_errp = dnde_coord[ij, ind, ik, il] - dnde

    ind = np.abs(stat_diff + invalid_value * (~maskp) - n_sigma_ul**2).argmin(axis=1)
    dnde_ul = dnde_coord[ij, ind, ik, il]

    dnde_err = (dnde_errn + dnde_errp) / 2

    maps = dict(
        ts=ts,
        dnde=dnde,
        dnde_err=dnde_err,
        dnde_errn=dnde_errn,
        dnde_errp=dnde_errp,
        dnde_ul=dnde_ul,
    )
    for key in maps.keys():
        maps[key] = Map.from_geom(geom, data=maps[key].value, unit=maps[key].unit)
    kwargs = dict(sed_type="dnde", gti=gti, reference_model=reference_model, meta=meta)
    output_maps.update(maps)
    return FluxMaps.from_maps(output_maps, **kwargs)


def _generate_scan_values(power_min=-6, power_max=2, relative_error=1e-2):
    """Values sampled such as we can probe a given `relative_error` on the norm
    between 10**`power_min` and 10**`power_max`.

    """
    arrays = []
    for power in range(power_min, power_max):
        vmin = 10**power
        vmax = 10 ** (power + 1)
        bin_per_decade = int((vmax - vmin) / (vmin * relative_error))
        arrays.append(np.linspace(vmin, vmax, bin_per_decade + 1, dtype=np.float32))
    scan_1side = np.unique(np.concatenate(arrays))
    return np.concatenate((-scan_1side[::-1], [0], scan_1side))


def _get_default_norm(
    norm,
    scan_min=0.2,
    scan_max=5,
    scan_n_values=11,
    scan_n_sigma=2,
    scan_values=None,
    interp="lin",
):
    """Create default norm parameter."""
    if norm is None or isinstance(norm, dict):
        norm_kwargs = dict(
            name="norm",
            value=1,
            unit="",
            interp=interp,
            frozen=False,
            scan_min=scan_min,
            scan_max=scan_max,
            scan_n_values=scan_n_values,
            scan_n_sigma=scan_n_sigma,
            scan_values=scan_values,
        )
        if isinstance(norm, dict):
            norm_kwargs.update(norm)
        try:
            norm = Parameter(**norm_kwargs)
        except TypeError as error:
            raise TypeError(f"Invalid dict key for norm init : {error}")
    if norm.name != "norm":
        raise ValueError("norm.name is not 'norm'")
    return norm


def _get_norm_scan_values(norm, result):
    """Compute norms based on the fit result to sample the stat profile at different scales."""
    norm_err = result["norm_err"]
    norm_value = result["norm"]
    if ~np.isfinite(norm_err) or norm_err == 0:
        norm_err = 0.1
    if ~np.isfinite(norm_value) or norm_value == 0:
        norm_value = 1.0
    sparse_norms = np.concatenate(
        (
            norm_value + np.linspace(-2.5, 2.5, 51) * norm_err,
            norm_value + np.linspace(-10, 10, 21) * norm_err,
            np.abs(norm_value) * np.linspace(-10, 10, 21),
            np.linspace(-10, 10, 21),
            np.linspace(norm.scan_values[0], norm.scan_values[-1], 2),
        )
    )
    sparse_norms = np.unique(sparse_norms)
    if len(sparse_norms) != 109:
        rand_norms = 20 * np.random.rand(109 - len(sparse_norms)) - 10
        sparse_norms = np.concatenate((sparse_norms, rand_norms))
    return np.sort(sparse_norms)


def apply_threshold_sensitivity(
    background, excess_counts, gamma_min=10, bkg_syst_fraction=0.05
):
    """Apply sensitivity  threshold in case it is limited by statistic or background."""
    is_gamma_limited = excess_counts < gamma_min
    excess_counts[is_gamma_limited] = gamma_min
    bkg_syst_limited = excess_counts < bkg_syst_fraction * background
    excess_counts[bkg_syst_limited] = bkg_syst_fraction * background[bkg_syst_limited]
    return excess_counts
